"""
Typing definitions for thelper.
"""

import io

import matplotlib.pyplot as plt
import numpy as np
import torch

from typing import (TYPE_CHECKING, Any, AnyStr, Callable, Dict,  # isort:skip
                    Generic, Iterable, List, Optional, Tuple, TypeVar, Union)  # noqa: F401

if TYPE_CHECKING:
    from thelper.tasks.utils import Task
    from thelper.nn.utils import Module
    from thelper.data.loaders import DataLoader
    from thelper.tasks.detect import BoundingBox

    ArrayType = np.ndarray  # generic definition
    ArrayShapeType = Union[List[int], Tuple[int]]
    OneOrManyArrayType = Union[List[ArrayType], ArrayType]

    # FIXME: experimental
    #  more verbatim than 'Any' or plain 'ndarray', but could flag invalid types in come places
    # more explicit definition, ex: ArrayOfType[myClass, ArrayShape[10, 10]]
    # more examples: https://github.com/python/typing/issues/516#issuecomment-350745218
    ArrayDType = TypeVar('ArrayDType', bound=np.generic)
    ArrayShape = TypeVar('ArrayShape', bound=Iterable[int])
    T = TypeVar('T')

    class ndarray(Generic[ArrayDType, ArrayShape]):  # noqa: E306  # name must match original exactly
        @property
        def shape(self):    # type: (...) -> Union[Tuple[int], Tuple[()]]
            return ()

    ArrayOfType = ndarray

    # specialized arrays
    W = TypeVar('W', bound=int)
    H = TypeVar('H', bound=int)
    C = TypeVar('C', bound=int)
    BoundingBoxArray = ArrayOfType[BoundingBox]
    ImageArray = ArrayOfType[ArrayType, ArrayShape[W, H, C]]

    ClassIdType = Union[AnyStr, int]
    LabelColorMapType = Union[ArrayType, Dict[int, ArrayType]]
    LabelIndex = AnyStr
    LabelType = AnyStr
    LabelDict = Dict[LabelIndex, LabelType]
    LabelList = List[LabelType]
    DrawingType = Optional[Tuple[plt.Figure, plt.Axes]]
    ClassColorMap = Dict[ClassIdType, Union[int, Tuple[int, int, int]]]

    Number = Union[int, float]
    _literalJSON = Optional[Union[AnyStr, Number, bool]]
    JSON = Union[_literalJSON, List[Union[_literalJSON, "JSON"]], Dict[AnyStr, Union[_literalJSON, "JSON"]]]

    SampleType = Dict[Union[AnyStr, int], Any]
    InputType = torch.Tensor
    ClassificationPredictionType = torch.Tensor
    ClassificationTargetType = torch.Tensor
    DetectionPredictionType = List[List[BoundingBox]]
    DetectionTargetType = List[List[BoundingBox]]
    AnyPredictionType = Union[ClassificationPredictionType, DetectionPredictionType]
    AnyTargetType = Union[ClassificationTargetType, DetectionTargetType]

    ConfigIndex = AnyStr
    ConfigValue = Union[AnyStr, bool, float, int, List[Any], Dict[Any, Any]]
    ConfigDict = Dict[ConfigIndex, Union[ConfigValue, "ConfigDict"]]

    CheckpointLoadingType = Union[AnyStr, io.FileIO]
    CheckpointContentType = Dict[AnyStr, Any]
    MapLocationType = Union[Callable, AnyStr, Dict[AnyStr, AnyStr]]

    ModelType = Module
    LoaderType = DataLoader
    MultiLoaderType = Tuple[Optional[LoaderType], Optional[LoaderType], Optional[LoaderType]]

    IterCallbackType = Callable[[Task, InputType, AnyPredictionType, AnyTargetType,
                                SampleType, Optional[float], int, int, int, int], None]
else:
    IterCallbackParams = [
        "task",         # the task object that defines class names, min/max target values, etc.
        "input",        # the (batched) input tensor given to the model in order to generate a prediction
        "pred",         # the (batched) tensor generated by the model containing predicted value(s)
        "target",       # the (batched) tensor containing target (groundtruth) prediction value(s)
        "sample",       # the mini-batch sample dictionary assembled by the data loader
        "loss",         # the loss computed by the model for the current iteration (may be None)
        "iter_idx",     # the index of the iteration (or sample index) in the current epoch
        "max_iters",    # the total number of iterations in the current epoch
        "epoch_idx",    # the index of the current epoch
        "max_epochs",   # the total (maximum) number of epochs the model should be trained for
    ]



"""Data parsers & utilities module for OGC-related projects."""

import copy
import functools
import logging

import cv2 as cv
import numpy as np
import tqdm

import thelper.data
import thelper.data.geo as geo
import thelper.train.utils

logger = logging.getLogger(__name__)


class TB15D104:
    """Wrapper class for OGC Testbed-15 (D104) identifiers."""

    TYPECE_RIVER = "10"
    TYPECE_LAKE = "21"

    BACKGROUND_ID = 0
    LAKE_ID = 1


class TB15D104Dataset(geo.parsers.VectorCropDataset):
    """OGC Testbed-15 dataset parser for D104 (lake/river) segmentation task."""

    def __init__(self, raster_path, vector_path, px_size=None,
                 allow_outlying_vectors=True, clip_outlying_vectors=True,
                 lake_area_min=0.0, lake_area_max=float("inf"),
                 lake_river_max_dist=float("inf"), feature_buffer=1000,
                 master_roi=None, focus_lakes=True, srs_target="3857", force_parse=False,
                 reproj_rasters=False, reproj_all_cpus=True, display_debug=False,
                 keep_rasters_open=True, parallel=False, transforms=None):
        assert isinstance(lake_river_max_dist, (float, int)) and lake_river_max_dist >= 0, "unexpected dist type"
        self.lake_river_max_dist = float(lake_river_max_dist)
        assert isinstance(focus_lakes, bool), "unexpected flag type"
        self.focus_lakes = focus_lakes
        assert px_size is None or isinstance(px_size, (float, int)), "pixel size (resolution) must be float/int"
        px_size = (1.0, 1.0) if px_size is None else (float(px_size), float(px_size))
        # note: we wrap partial static functions for caching to see when internal parameters are changing
        cleaner = functools.partial(self.lake_cleaner, area_min=lake_area_min, area_max=lake_area_max,
                                    lake_river_max_dist=lake_river_max_dist, parallel=parallel)
        if self.focus_lakes:
            cropper = functools.partial(self.lake_cropper, px_size=px_size, skew=(0.0, 0.0),
                                        feature_buffer=feature_buffer, parallel=parallel)
        else:
            # TODO: implement river-focused cropper (i.e. river-length parsing?)
            raise NotImplementedError
        super().__init__(raster_path=raster_path, vector_path=vector_path, px_size=px_size, skew=None,
                         allow_outlying_vectors=allow_outlying_vectors, clip_outlying_vectors=clip_outlying_vectors,
                         vector_area_min=lake_area_min, vector_area_max=lake_area_max, vector_target_prop=None,
                         feature_buffer=feature_buffer, master_roi=master_roi, srs_target=srs_target,
                         raster_key="lidar", mask_key="hydro", cleaner=cleaner, cropper=cropper,
                         force_parse=force_parse, reproj_rasters=reproj_rasters, reproj_all_cpus=reproj_all_cpus,
                         keep_rasters_open=keep_rasters_open, transforms=transforms)
        meta_keys = self.task.meta_keys
        if "bboxes" in meta_keys:
            del meta_keys[meta_keys.index("bboxes")]  # placed in meta list by base class constr, moved to detect target below
        self.task = thelper.tasks.Detection(class_names={"background": TB15D104.BACKGROUND_ID, "lake": TB15D104.LAKE_ID},
                                            input_key="input", bboxes_key="bboxes",
                                            meta_keys=meta_keys, background=0, color_map={"lake": [255, 0, 0]})
        # update all already-created bboxes with new task ref
        for s in self.samples:
            for b in s["bboxes"]:
                b.task = self.task
        self.display_debug = display_debug
        self.parallel = parallel

    @staticmethod
    def lake_cleaner(features, area_min, area_max, lake_river_max_dist, parallel=False):
        """Flags geometric features as 'clean' based on type and distance to nearest river."""
        # note: we use a flag here instead of removing bad features so that end-users can still use them if needed
        for f in features:
            f["clean"] = False  # flag every as 'bad' by default, clear just the ones of interest below
        rivers = [f for f in features if f["properties"]["TYPECE"] == TB15D104.TYPECE_RIVER]
        lakes = [f for f in features if f["properties"]["TYPECE"] == TB15D104.TYPECE_LAKE]
        logger.info(f"labeling and cleaning {len(lakes)} lakes...")

        def clean_lake(lake):
            if area_min <= lake["geometry"].area <= area_max:
                if lake_river_max_dist == float("inf"):
                    return True
                else:
                    for river in rivers:
                        # note: distance check below seems to be "intelligent", i.e. it will
                        # first check bbox distance, then check chull distance, and finally use
                        # the full geometries (doing these steps here explicitly makes it slower)
                        if lake["geometry"].distance(river["geometry"]) < lake_river_max_dist:
                            return True
            return False

        if parallel:
            if not isinstance(parallel, int):
                import multiprocessing
                parallel = multiprocessing.cpu_count()
            assert parallel > 0, "unexpected min core count"
            import joblib
            flags = joblib.Parallel(n_jobs=parallel)(joblib.delayed(
                clean_lake)(lake) for lake in tqdm.tqdm(lakes, desc="labeling + cleaning lakes"))
            for flag, lake in zip(flags, lakes):
                lake["clean"] = flag
        else:
            for lake in tqdm.tqdm(lakes, desc="labeling + cleaning lakes"):
                lake["clean"] = clean_lake(lake)
        return features

    @staticmethod
    def lake_cropper(features, rasters_data, coverage, srs_target, px_size, skew, feature_buffer, parallel=False):
        """Returns the ROI information for a given feature (may be modified in derived classes)."""
        srs_target_wkt = srs_target.ExportToWkt()

        def crop_feature(feature):
            assert feature["clean"]  # should not get here with bad features
            roi, roi_tl, roi_br, crop_width, crop_height = \
                geo.utils.get_feature_roi(feature["geometry"], px_size, skew, feature_buffer)
            roi_geotransform = (roi_tl[0], px_size[0], skew[0],
                                roi_tl[1], skew[1], px_size[1])
            # test all raster regions that touch the selected feature
            raster_hits = []
            for raster_idx, raster_data in enumerate(rasters_data):
                if raster_data["target_roi"].intersects(roi):
                    raster_hits.append(raster_idx)
            # make list of all other features that may be included in the roi
            roi_centroid = feature["centroid"]
            roi_radius = np.linalg.norm(np.asarray(roi_tl) - np.asarray(roi_br)) / 2
            roi_features, bboxes = [], []
            # note: the 'image id' is in fact the id of the focal feature in the crop
            image_id = int(feature["properties"]["OBJECTID"])
            for f in features:
                # note: here, f may not be 'clean', test anyway
                if f["geometry"].distance(roi_centroid) > roi_radius:
                    continue
                inters = f["geometry"].intersection(roi)
                if inters.is_empty:
                    continue
                roi_features.append(f)
                if f["properties"]["TYPECE"] == TB15D104.TYPECE_RIVER:
                    continue
                # only lakes can generate bboxes; make sure to clip them to the roi bounds
                clip = f["clipped"] or not inters.equals(f["geometry"])
                if clip:
                    assert inters.geom_type in ["Polygon", "MultiPolygon"], "unexpected inters type"
                    corners = []
                    if inters.geom_type == "Polygon":
                        bounds = inters.bounds
                        corners.append((bounds[0:2], bounds[2:4]))
                    elif inters.geom_type == "MultiPolygon":
                        for poly in inters:
                            bounds = poly.bounds
                            corners.append((bounds[0:2], bounds[2:4]))
                else:
                    corners = [(f["tl"], f["br"])]
                for c in corners:
                    feat_tl_px = geo.utils.get_pxcoord(roi_geotransform, *c[0])
                    feat_br_px = geo.utils.get_pxcoord(roi_geotransform, *c[1])
                    bbox = [max(0, feat_tl_px[0]), max(0, feat_tl_px[1]),
                            min(crop_width - 1, feat_br_px[0]),
                            min(crop_height - 1, feat_br_px[1])]
                    if bbox[2] - bbox[0] <= 1 or bbox[3] - bbox[1] <= 1:
                        continue  # skip all bboxes smaller than 1 px (c'mon...)
                    # note: lake class id is 1 by definition
                    bboxes.append(thelper.tasks.detect.BoundingBox(TB15D104.LAKE_ID,
                                                                   bbox=bbox,
                                                                   include_margin=False,
                                                                   truncated=clip,
                                                                   image_id=image_id))
            # prepare actual 'sample' for crop generation at runtime
            return {
                "features": roi_features,
                "bboxes": bboxes,
                "focal": feature,
                "id": image_id,
                "roi": roi,
                "roi_tl": roi_tl,
                "roi_br": roi_br,
                "raster_hits": raster_hits,
                "crop_width": crop_width,
                "crop_height": crop_height,
                "geotransform": np.asarray(roi_geotransform),
                "srs": srs_target_wkt,
            }

        clean_feats = [f for f in features if f["clean"]]
        if parallel:
            if not isinstance(parallel, int):
                import multiprocessing
                parallel = multiprocessing.cpu_count()
            assert parallel > 0, "unexpected min core count"
            import joblib
            samples = joblib.Parallel(n_jobs=parallel)(joblib.delayed(
                crop_feature)(feat) for feat in tqdm.tqdm(clean_feats, desc="preparing crop regions"))
        else:
            samples = []
            for feature in tqdm.tqdm(clean_feats, desc="validating crop candidates"):
                samples.append(crop_feature(feature))
        return [s for s in samples if s is not None]

    def _show_stats_plots(self, show=False, block=False):
        """Draws and returns feature stats histograms using pyplot."""
        import matplotlib.pyplot as plt
        feature_categories = {}
        for feat in self.features:
            curr_cat = feat["properties"]["TYPECE"]
            if curr_cat not in feature_categories:
                feature_categories[curr_cat] = []
            feature_categories[curr_cat].append(feat)
        fig, axes = plt.subplots(len(feature_categories))
        for idx, (cat, features) in enumerate(feature_categories.items()):
            areas = [f["geometry"].area for f in features]
            axes[idx].hist(areas, density=True, bins=30,
                           range=(max(self.area_min, min(areas)), min(self.area_max, max(areas))))
            axes[idx].set_xlabel("Surface (m^2)")
            axes[idx].set_title(f"TYPECE = {cat}")
            axes[idx].set_xlim(xmin=0)
        if show:
            fig.show()
            if block:
                plt.show(block=block)
                return fig
            plt.pause(0.5)
        return fig, axes

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        assert idx < len(self.samples), "sample index is out-of-range"
        if idx < 0:
            idx = len(self.samples) + idx
        sample = self.samples[idx]
        crop, mask = self._process_crop(sample)
        assert crop.shape[2] == 1, "unexpected lidar raster band count"
        crop = crop[:, :, 0]
        dmap = cv.distanceTransform(np.where(mask, np.uint8(0), np.uint8(255)), cv.DIST_L2, cv.DIST_MASK_PRECISE)
        dmap_inv = cv.distanceTransform(np.where(mask, np.uint8(255), np.uint8(0)), cv.DIST_L2, cv.DIST_MASK_PRECISE)
        dmap = np.where(mask, -dmap_inv, dmap)
        dmap *= self.px_size[0]  # constructor enforces same px width/height size
        # dc mask is crop.mask, but most likely lost below
        if self.display_debug:
            crop = cv.normalize(crop, dst=crop, alpha=0, beta=255, norm_type=cv.NORM_MINMAX,
                                dtype=cv.CV_8U, mask=(~crop.mask).astype(np.uint8))
            mask = cv.normalize(mask, dst=mask, alpha=0, beta=255, norm_type=cv.NORM_MINMAX, dtype=cv.CV_8U)
            dmap = cv.normalize(dmap, dst=dmap, alpha=0, beta=255, norm_type=cv.NORM_MINMAX, dtype=cv.CV_8U)
        sample = {
            "input": np.stack([crop, mask, dmap], axis=-1),
            # note: bboxes are automatically added in the "cropper" preprocessing function
            "hydro": mask,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class TB15D104TileDataset(geo.parsers.TileDataset):
    """OGC Testbed-15 dataset parser for D104 (lake/river) segmentation task."""

    def __init__(self, raster_path, vector_path, tile_size, tile_overlap,
                 px_size=None, allow_outlying_vectors=True, clip_outlying_vectors=True,
                 lake_area_min=0.0, lake_area_max=float("inf"), master_roi=None, srs_target="3857",
                 force_parse=False, reproj_rasters=False, reproj_all_cpus=True, display_debug=False,
                 keep_rasters_open=True, parallel=False, transforms=None):
        assert px_size is None or isinstance(px_size, (float, int)), "pixel size (resolution) must be float/int"
        px_size = (1.0, 1.0) if px_size is None else (float(px_size), float(px_size))
        # note: we wrap partial static functions for caching to see when internal parameters are changing
        cleaner = functools.partial(TB15D104Dataset.lake_cleaner, area_min=lake_area_min, area_max=lake_area_max,
                                    lake_river_max_dist=float("inf"), parallel=parallel)
        super().__init__(raster_path=raster_path, vector_path=vector_path, tile_size=tile_size,
                         tile_overlap=tile_overlap, skip_empty_tiles=True, skip_nodata_tiles=False,
                         px_size=px_size, allow_outlying_vectors=allow_outlying_vectors,
                         clip_outlying_vectors=clip_outlying_vectors, vector_area_min=lake_area_min,
                         vector_area_max=lake_area_max, vector_target_prop=None, master_roi=master_roi,
                         srs_target=srs_target, raster_key="lidar", mask_key="hydro", cleaner=cleaner,
                         force_parse=force_parse, reproj_rasters=reproj_rasters, reproj_all_cpus=reproj_all_cpus,
                         keep_rasters_open=keep_rasters_open, transforms=transforms)
        meta_keys = self.task.meta_keys
        self.task = thelper.tasks.Detection(class_names={"background": TB15D104.BACKGROUND_ID, "lake": TB15D104.LAKE_ID},
                                            input_key="input", bboxes_key="bboxes",
                                            meta_keys=meta_keys, background=0, color_map={"lake": [255, 0, 0]})
        self.display_debug = display_debug
        self.parallel = parallel

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        assert idx < len(self.samples), "sample index is out-of-range"
        if idx < 0:
            idx = len(self.samples) + idx
        sample = self.samples[idx]
        crop, mask = self._process_crop(sample)
        assert crop.shape[2] == 1, "unexpected lidar raster band count"
        crop = crop[:, :, 0]
        dmap = cv.distanceTransform(np.where(mask, np.uint8(0), np.uint8(255)), cv.DIST_L2, cv.DIST_MASK_PRECISE)
        dmap_inv = cv.distanceTransform(np.where(mask, np.uint8(255), np.uint8(0)), cv.DIST_L2, cv.DIST_MASK_PRECISE)
        dmap = np.where(mask, -dmap_inv, dmap)
        dmap *= self.px_size[0]  # constructor enforces same px width/height size
        # dc mask is crop.mask, but most likely lost below
        if self.display_debug:
            crop = cv.normalize(crop, dst=crop, alpha=0, beta=255, norm_type=cv.NORM_MINMAX,
                                dtype=cv.CV_8U, mask=(~crop.mask).astype(np.uint8))
            mask = cv.normalize(mask, dst=mask, alpha=0, beta=255, norm_type=cv.NORM_MINMAX, dtype=cv.CV_8U)
            dmap = cv.normalize(dmap, dst=dmap, alpha=0, beta=255, norm_type=cv.NORM_MINMAX, dtype=cv.CV_8U)
        # note1: contrarily to the 'smart' dataset parser above, we need to add bboxes here
        # note2: bboxes should only be added over areas that are not 'nodata'
        bboxes = []
        for f in sample["features"]:
            if f["properties"]["TYPECE"] == TB15D104.TYPECE_RIVER:
                continue
            # only lakes can generate bboxes; make sure to clip them to the crop bounds
            inters = f["geometry"].intersection(sample["roi"])
            clip = f["clipped"] or not inters.equals(f["geometry"])
            if clip:
                assert inters.geom_type in ["Polygon", "MultiPolygon"], "unexpected inters type"
                corners = []
                if inters.geom_type == "Polygon":
                    bounds = inters.bounds
                    corners.append((bounds[0:2], bounds[2:4]))
                elif inters.geom_type == "MultiPolygon":
                    for poly in inters:
                        bounds = poly.bounds
                        corners.append((bounds[0:2], bounds[2:4]))
            else:
                corners = [(f["tl"], f["br"])]
            for c in corners:
                feat_tl_px = geo.utils.get_pxcoord(sample["geotransform"], *c[0])
                feat_br_px = geo.utils.get_pxcoord(sample["geotransform"], *c[1])
                bbox = [max(0, feat_tl_px[0]), max(0, feat_tl_px[1]),
                        min(sample["crop_width"] - 1, feat_br_px[0]),
                        min(sample["crop_height"] - 1, feat_br_px[1])]
                if bbox[2] - bbox[0] <= 1 or bbox[3] - bbox[1] <= 1:
                    continue  # skip all bboxes smaller than 1 px (c'mon...)
                # note: lake class id is 1 by definition
                bboxes.append(thelper.tasks.detect.BoundingBox(TB15D104.LAKE_ID,
                                                               bbox=bbox,
                                                               include_margin=False,
                                                               truncated=clip,
                                                               image_id=sample["id"],
                                                               task=self.task))
        sample = {
            "input": np.stack([crop, mask, dmap], axis=-1),
            "hydro": mask,
            "bboxes": bboxes,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class TB15D104DetectLogger(thelper.train.utils.DetectLogger):

    def __init__(self, conf_threshold=0.5):
        super().__init__(conf_threshold=conf_threshold, target_name="lake",
                         log_keys=["id", "geotransform"], format="geojson")

    def report_geojson(self):
        # here, we only care about reporting predictions, we ignore the (possibly missing) gt bboxes
        import shapely
        import geojson
        batch_size = len(self.bbox[0])
        bbox_lists = [bboxes for batch in self.bbox for bboxes in batch]  # one list per crop
        if batch_size > 1:
            geotransforms = [np.asarray(geot) for batch in self.meta["geotransform"] for geot in batch]
        else:
            # special check to avoid issues when unpacking (1,6)-dim tensor
            geotransforms = [np.asarray(geot) for geot in self.meta["geotransform"]]
        crop_ids = [id for batch in self.meta["id"] for id in batch]
        output_features = []
        for bboxes, geotransform, id in zip(bbox_lists, geotransforms, crop_ids):
            for bbox in bboxes:
                if geotransform.shape[0] == 1:
                    geotransform = geotransform[0]
                bbox_tl = geo.utils.get_geocoord(geotransform, *bbox.top_left)
                bbox_br = geo.utils.get_geocoord(geotransform, *bbox.bottom_right)
                bbox_geom = shapely.geometry.Polygon([bbox_tl, (bbox_br[0], bbox_tl[1]),
                                                      bbox_br, (bbox_tl[0], bbox_br[1])])
                output_features.append(geojson.Feature(geometry=bbox_geom, properties={
                    "image_id": id, "confidence": bbox.confidence}))
        return geojson.dumps(geojson.FeatureCollection(output_features), indent=2)


def postproc_features(input_file, bboxes_srs, orig_geoms_path, output_file,
                      final_srs=None, write_shapefile_copy=False):
    """Post-processes bounding box detections produced during an evaluation session into a GeoJSON file."""
    import ogr
    import osr
    import json
    import geojson
    import shapely
    logger.debug("importing bboxes SRS...")
    assert isinstance(bboxes_srs, (str, int, osr.SpatialReference)), \
        "target EPSG SRS must be given as int/str"
    if isinstance(bboxes_srs, (str, int)):
        if isinstance(bboxes_srs, str):
            bboxes_srs = int(bboxes_srs.replace("EPSG:", ""))
        bboxes_srs_obj = osr.SpatialReference()
        bboxes_srs_obj.ImportFromEPSG(bboxes_srs)
        bboxes_srs = bboxes_srs_obj
    logger.debug("importing lake bboxes geojson...")
    with open(input_file) as bboxes_fd:
        bboxes_geoms = thelper.data.geo.utils.parse_geojson(json.load(bboxes_fd))
    logger.debug("importing hydro features geojson...")
    with open(orig_geoms_path) as hydro_fd:
        hydro_geoms = thelper.data.geo.utils.parse_geojson(json.load(hydro_fd), srs_target=bboxes_srs)
    logger.debug("computing global cascade of lake bboxes...")
    detect_roi = shapely.ops.cascaded_union([bbox["geometry"] for bbox in bboxes_geoms])
    output_features = []

    def append_poly(feat, props, srs_transform=None):
        if feat.is_empty:
            return
        elif feat.type == "Polygon":
            if srs_transform is not None:
                ogr_geometry = ogr.CreateGeometryFromWkb(feat.wkb)
                ogr_geometry.Transform(srs_transform)
                feat = shapely.wkt.loads(ogr_geometry.ExportToWkt())
            output_features.append((geojson.Feature(geometry=feat, properties=props), feat))
        elif feat.type == "MultiPolygon" or feat.type == "GeometryCollection":
            for f in feat:
                append_poly(f, props)

    srs_transform = None
    if final_srs is not None:
        import osr
        logger.debug("importing output SRS...")
        assert isinstance(final_srs, (str, int, osr.SpatialReference)), \
            "target EPSG SRS must be given as int/str"
        if isinstance(final_srs, (str, int)):
            if isinstance(final_srs, str):
                final_srs = int(final_srs.replace("EPSG:", ""))
            final_srs_obj = osr.SpatialReference()
            final_srs_obj.ImportFromEPSG(final_srs)
            final_srs = final_srs_obj
        if not bboxes_srs.IsSame(final_srs):
            srs_transform = osr.CoordinateTransformation(bboxes_srs, final_srs)
    logger.debug("running hydro feature and lake bboxes intersection loop...")
    for hydro_feat in tqdm.tqdm(hydro_geoms, desc="computing bbox intersections"):
        # find intersection and append to list of 'lakes'
        intersection = hydro_feat["geometry"].intersection(detect_roi)
        hydro_feat["properties"]["TYPECE"] = TB15D104.TYPECE_LAKE
        append_poly(intersection, copy.deepcopy(hydro_feat["properties"]), srs_transform)
        if not intersection.is_empty:
            # subtract bbox region from feature if intersection found (leftovers at end will be 'rivers')
            hydro_feat["geometry"] = hydro_feat["geometry"].difference(detect_roi)
    logger.debug("running river cleanup loop...")
    for hydro_feat in tqdm.tqdm(hydro_geoms, desc="appending leftover geometries as rivers"):
        if not hydro_feat["geometry"].is_empty:
            # remark: hydro features outside the original ROI will appear as rivers despite never being processed
            hydro_feat["properties"]["TYPECE"] = TB15D104.TYPECE_RIVER
            append_poly(hydro_feat["geometry"], copy.deepcopy(hydro_feat["properties"]), srs_transform)
    logger.debug("exporting final geojson...")
    with open(output_file, "w") as fd:
        out_srs = final_srs if final_srs is not None else bboxes_srs
        fd.write(geo.utils.export_geojson_with_crs([o[0] for o in output_features], srs_target=out_srs))
    if write_shapefile_copy:
        logger.debug("exporting final shapefile...")
        driver = ogr.GetDriverByName("ESRI Shapefile")
        data_source = driver.CreateDataSource(output_file + ".shp")
        layer = data_source.CreateLayer("lakes", final_srs if final_srs is not None else bboxes_srs, ogr.wkbPolygon)
        for feat_tuple in output_features:
            feature = ogr.Feature(layer.GetLayerDefn())
            point = ogr.CreateGeometryFromWkt(feat_tuple[1].wkt)
            feature.SetGeometry(point)
            layer.CreateFeature(feature)
            feature = None
        data_source = None



"""Detection task interface module.

This module contains classes that define object detection utilities and task interfaces.
"""
import logging
from typing import List, Optional, Tuple, Union  # noqa: F401

import numpy as np
import torch
import tqdm

import thelper.concepts
import thelper.utils
from thelper.ifaces import ClassNamesHandler
from thelper.tasks.regr import Regression
from thelper.tasks.utils import Task

logger = logging.getLogger(__name__)


@thelper.concepts.detection
class BoundingBox:
    """Interface used to hold instance metadata for object detection tasks.

    Object detection trainers and display utilities in the framework will expect this interface to be
    used when parsing a predicted detection or an annotation. The base contents are based on the
    PASCALVOC metadata structure, and this class can be derived if necessary to contain more metadata.

    Attributes:
        class_id: type identifier for the underlying object instance.
        bbox: four-element tuple holding the (xmin,ymin,xmax,ymax) bounding box parameters.
        include_margin: defines whether xmax/ymax is included in the bounding box area or not.
        difficult: defines whether this instance is considered "difficult" (false by default).
        occluded: defines whether this instance is considered "occluded" (false by default).
        truncated: defines whether this instance is considered "truncated" (false by default).
        iscrowd: defines whether this instance covers a "crowd" of objects or not (false by default).
        confidence: scalar or array of prediction confidence values tied to class types (empty by default).
        image_id: string used to identify the image containing this bounding box (i.e. file path or uuid).
        task: reference to the task object that holds extra metadata regarding the content of the bbox (None by default).

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.tasks.detect.Detection`
    """

    def __init__(self, class_id, bbox, include_margin=True, difficult=False, occluded=False,
                 truncated=False, iscrowd=False, confidence=None, image_id=None, task=None):
        """Receives and stores low-level input detection metadata for later access."""
        self.class_id = class_id  # should be string or int to allow batching in data loaders
        # note: the input bbox is expected to be a 4 element array (xmin,ymin,xmax,ymax)
        self.include_margin = include_margin
        self.bbox = bbox
        self.difficult = difficult
        self.occluded = occluded
        self.truncated = truncated
        self.iscrowd = iscrowd
        self.confidence = confidence
        self.image_id = image_id  # should be string that identifies the associated image (file path or uuid)
        self.task = task

    @property
    def class_id(self):
        # type: () -> Union[int, str]
        """Returns the object class type identifier."""
        return self._class_id

    @class_id.setter
    def class_id(self, value):
        # type: (Union[int, str]) -> None
        """Sets the object class type identifier  (should be string/int)."""
        assert isinstance(value, (int, str)), "class should be defined as integer (index) or string (name)"
        self._class_id = value

    @property
    def bbox(self):
        """Returns the bounding box tuple :math:`(x_min,y_min,x_max,y_max)`."""
        return self._bbox

    @bbox.setter
    def bbox(self, value):
        """Sets the bounding box tuple :math:`(x_min,y_min,x_max,y_max)`."""
        assert isinstance(value, (list, tuple, np.ndarray, torch.Tensor)) and len(value) == 4, "invalid input type/len"
        assert not isinstance(value, (list, tuple)) or all([isinstance(v, (int, float)) for v in value]), \
            "input bbox values must be integer/float"
        assert value[0] <= value[2] and value[1] <= value[3], "invalid min/max values for bbox coordinates"
        assert not self.include_margin or not any([isinstance(v, float) for v in value]), \
            "it makes no sense to include xmax/ymax margin if using floating point coordinates"
        self._bbox = value

    @property
    def left(self):
        """Returns the left bounding box edge origin offset value."""
        return self._bbox[0]

    @left.setter
    def left(self, value):
        """Sets the left bounding box edge origin offset value."""
        self._bbox[0] = value

    @property
    def top(self):
        """Returns the top bounding box edge origin offset value."""
        return self._bbox[1]

    @top.setter
    def top(self, value):
        """Sets the top bounding box edge origin offset value."""
        self._bbox[1] = value

    @property
    def top_left(self):
        """Returns the top left bounding box corner coordinates :math:`(x,y)`."""
        return self._bbox[0], self._bbox[1]

    @top_left.setter
    def top_left(self, value):
        """Sets the top left bounding box corner coordinates :math:`(x,y)`."""
        assert isinstance(value, (list, tuple, np.ndarray, torch.Tensor)) and len(value) == 2, "invalid input type/len"
        self._bbox[0], self._bbox[1] = value[0], value[1]

    @property
    def right(self):
        """Returns the right bounding box edge origin offset value."""
        return self._bbox[2]

    @right.setter
    def right(self, value):
        """Sets the right bounding box edge origin offset value."""
        self._bbox[2] = value

    @property
    def bottom(self):
        """Returns the bottom bounding box edge origin offset value."""
        return self._bbox[3]

    @bottom.setter
    def bottom(self, value):
        """Sets the bottom bounding box edge origin offset value."""
        self._bbox[3] = value

    @property
    def bottom_right(self):
        """Returns the bottom right bounding box corner coordinates :math:`(x,y)`."""
        return self._bbox[2], self._bbox[3]

    @bottom_right.setter
    def bottom_right(self, value):
        """Sets the bottom right bounding box corner coordinates :math:`(x,y)`."""
        assert isinstance(value, (list, tuple, np.ndarray, torch.Tensor)) and len(value) == 2, "invalid input type/len"
        assert value[0] >= self._bbox[0] and value[1] >= self._bbox[1]
        self._bbox[2], self._bbox[3] = value[0], value[1]

    @property
    def width(self):
        """Returns the width of the bounding box."""
        return (self._bbox[2] - self._bbox[0]) + 1 if self.include_margin else 0

    @property
    def height(self):
        """Returns the height of the bounding box."""
        return (self._bbox[3] - self._bbox[1]) + 1 if self.include_margin else 0

    @property
    def centroid(self, floor=False):
        """Returns the bounding box centroid coordinates :math:`(x,y)`."""
        if self.include_margin:
            if floor:
                return (self._bbox[0] + self._bbox[2] + 1) // 2, (self._bbox[1] + self._bbox[3] + 1) // 2
            return (self._bbox[0] + self._bbox[2] + 1) / 2, (self._bbox[1] + self._bbox[3] + 1) / 2
        else:
            if floor:
                return (self._bbox[0] + self._bbox[2]) // 2, (self._bbox[1] + self._bbox[3]) // 2
            return (self._bbox[0] + self._bbox[2]) / 2, (self._bbox[1] + self._bbox[3]) / 2

    @property
    def include_margin(self):
        """Returns whether :math:`x_max` and :math:`y_max` are included in the bounding box area or not"""
        return self._include_margin

    @include_margin.setter
    def include_margin(self, value):
        """Sets whether :math:`x_max` and :math:`y_max` are is included in the bounding box area or not"""
        assert isinstance(value, (int, bool)), "flag type must be integer or boolean"
        self._include_margin = value

    @property
    def difficult(self):
        """Returns whether this bounding box is considered *difficult* by the dataset (false by default)."""
        return self._difficult

    @difficult.setter
    def difficult(self, value):
        """Sets whether this bounding box is considered *difficult* by the dataset."""
        assert isinstance(value, (int, bool)), "flag type must be integer or boolean"
        self._difficult = value

    @property
    def occluded(self):
        """Returns whether this bounding box is considered *occluded* by the dataset (false by default)."""
        return self._occluded

    @occluded.setter
    def occluded(self, value):
        """Sets whether this bounding box is considered *occluded* by the dataset."""
        assert isinstance(value, (int, bool)), "flag type must be integer or boolean"
        self._occluded = value

    @property
    def truncated(self):
        """Returns whether this bounding box is considered *truncated* by the dataset (false by default)."""
        return self._truncated

    @truncated.setter
    def truncated(self, value):
        """Sets whether this bounding box is considered *truncated* by the dataset."""
        assert isinstance(value, (int, bool)), "flag type must be integer or boolean"
        self._truncated = value

    @property
    def iscrowd(self):
        """Returns whether this instance covers a *crowd* of objects or not (false by default)."""
        return self._iscrowd

    @iscrowd.setter
    def iscrowd(self, value):
        """Sets whether this instance covers a *crowd* of objects or not."""
        assert isinstance(value, (int, bool)), "flag type must be integer or boolean"
        self._iscrowd = value

    @property
    def area(self):
        """Returns a scalar indicating the total surface of the annotation (may be None if unknown/unspecified)."""
        return self.width * self.height

    @property
    def confidence(self):
        """Returns the confidence value (or array of confidence values) associated to the predicted class types."""
        return self._confidence

    @confidence.setter
    def confidence(self, value):
        """Sets the confidence value (or array of confidence values) associated to the predicted class types."""
        assert value is None or isinstance(value, (float, list, np.ndarray, torch.Tensor)), "value should be float/list/ndarray/tensor"
        self._confidence = value

    @property
    def image_id(self):
        """Returns the image string identifier."""
        return self._image_id

    @image_id.setter
    def image_id(self, value):
        """Sets the image string identifier."""
        assert value is None or isinstance(value, (str, int)), "image identifier should be a string/int (file path or uuid)"
        self._image_id = value

    @property
    def task(self):
        """Returns the reference to the task object that holds extra metadata regarding the content of the bbox."""
        return self._task

    @task.setter
    def task(self, value):
        """Sets the reference to the task object that holds extra metadata regarding the content of the bbox."""
        if value is not None:
            assert isinstance(value, Detection), "task should be detection-related"
            assert self.class_id in value.class_indices.values(), f"cannot find class_id '{self.class_id}' in task indices"
        self._task = value

    def encode(self, format=None):
        """Returns a vectorizable representation of this bounding box in a specified format.

        WARNING: Encoding might cause information loss (e.g. task reference is discarded).
        """
        if format == "coco":
            return [*self.top_left, self.width, self.height, self.class_id]
        elif format == "pascal_voc":
            return [*self.top_left, *self.bottom_right, self.class_id,
                    self.difficult, self.occluded, self.truncated]
        else:
            assert format is None, "unrecognized/unknown encoding format"
            vec = [*self.bbox, self.class_id, self.include_margin, self.difficult, self.occluded,
                   self.truncated, self.iscrowd, self.image_id]
            if self.confidence is not None:
                vec += [self.confidence] if isinstance(self.confidence, float) else [*self.confidence]
            return vec

    @staticmethod
    def decode(vec, format=None):
        """Returns a BoundingBox object from a vectorized representation in a specified format.

        .. note::
            The input bbox is expected to be a 4 element array :math:`(x_min,y_min,x_max,y_max)`.
        """
        if format == "coco":
            assert len(vec) == 5, "unexpected vector length (should contain 5 values)"
            return BoundingBox(class_id=vec[4], bbox=[vec[0], vec[1], vec[0] + vec[2], vec[1] + vec[3]])
        elif format == "pascal_voc":
            assert len(vec) == 8, "unexpected vector length (should contain 8 values)"
            return BoundingBox(class_id=vec[4], bbox=vec[0:4], difficult=vec[5], occluded=vec[6], truncated=vec[7])
        else:
            assert format is None, "unrecognized/unknown encoding format"
            assert len(vec) >= 12, "unexpected vector length (should contain 9 values or more)"
            return BoundingBox(class_id=vec[4], bbox=vec[0:4], include_margin=vec[5], difficult=vec[6],
                               occluded=vec[7], truncated=vec[8], iscrowd=vec[9],
                               confidence=(None if len(vec) == 11 else vec[11:]),
                               image_id=vec[10], task=None)

    def intersects(self, geom):
        """Returns whether the bounding box intersects a geometry (i.e. a 2D point or another bbox)."""
        assert isinstance(geom, (tuple, list, np.ndarray, BoundingBox)), "unexpected input geometry type"
        if isinstance(geom, (tuple, list, np.ndarray)):
            # check intersection with point
            assert len(geom) == 2, "point should be given as list of coordinates (x,y)"
            return self._bbox[0] <= geom[0] <= self.bbox[2] and self._bbox[1] <= geom[1] <= self.bbox[3]
        else:
            return not (self._bbox[0] > geom._bbox[2] or geom._bbox[0] > self._bbox[2] or
                        self._bbox[3] < geom._bbox[1] or geom._bbox[3] < self._bbox[1])

    def totuple(self):
        # type: () -> Tuple[int, int, int, int]
        """Gets a ``tuple`` representation of the underlying bounding box tuple :math:`(x_min,y_min,x_max,y_max)`.

        This ensures that ``Tensor`` objects are converted to native *Python* types."""
        return tuple(self.tolist())

    def tolist(self):
        # type: () -> List[int]
        """Gets a ``list`` representation of the underlying bounding box tuple :math:`(x_min,y_min,x_max,y_max)`.

        This ensures that ``Tensor`` objects are converted to native *Python* types."""
        return self._bbox.tolist() if isinstance(self._bbox, torch.Tensor) else list(self._bbox)

    def json(self):
        # type: () -> thelper.typedefs.JSON
        """Gets a JSON-serializable representation of the bounding box parameters."""
        return {
            "class_id": self.class_id,
            "image_id": self.image_id,
            "bbox": self.tolist(),
            "confidence": self.confidence,
            "include_margin": self.include_margin,
            "difficult": self.difficult,
            "occluded": self.occluded,
            "truncated": self.truncated,
            "is_crowd": self.iscrowd,
        }

    def __repr__(self):
        """Creates a print-friendly representation of the object detection bbox instance."""
        # note: we do not export the task reference here (it might be too heavy for logs)
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(class_id={repr(self.class_id)}, bbox={repr(self.bbox)}, include_margin={repr(self.include_margin)}, " + \
            f"difficult={repr(self.difficult)}, occluded={repr(self.occluded)}, truncated={repr(self.truncated)}, " + \
            f"iscrowd={repr(self.iscrowd)}, confidence={repr(self.confidence)}, image_id={repr(self.image_id)})"


@thelper.concepts.detection
class Detection(Regression, ClassNamesHandler):
    """Interface for object detection tasks.

    This specialization requests that when given an input image, the trained model should
    provide a list of bounding box (bbox) proposals that correspond to probable objects detected
    in the image.

    This specialized regression interface is currently used to help display functions.

    Attributes:
        class_names: map of class name-value pairs for object types to detect.
        input_key: the key used to fetch input tensors from a sample dictionary.
        bboxes_key: the key used to fetch target (groundtruth) bboxes from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.
        input_shape: a numpy-compatible shape to expect input images to possess.
        target_shape: a numpy-compatible shape to expect the predictions to be in.
        target_min: a 2-dim tensor containing minimum (x,y) bounding box corner values.
        target_max: a 2-dim tensor containing maximum (x,y) bounding box corner values.
        background: value of the 'background' label (if any) used in the class map.
        color_map: map of class name-color pairs to use when displaying results.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.tasks.regr.Regression`
        | :class:`thelper.train.detect.ObjDetectTrainer`
    """

    def __init__(self, class_names, input_key, bboxes_key, meta_keys=None, input_shape=None,
                 target_shape=None, target_min=None, target_max=None, background=None, color_map=None):
        """Receives and stores the bbox types to detect, the input tensor key, the groundtruth
        bboxes list key, the extra (meta) keys produced by the dataset parser(s), and the color
        map used to color bboxes when displaying results.

        The class names can be provided as a list of strings, as a path to a json file that
        contains such a list, or as a map of predefined name-value pairs to use in gt maps.
        This list/map must contain at least two elements (background and one class). All
        other arguments are used as-is to index dictionaries, and must therefore be key-
        compatible types.
        """
        super(Detection, self).__init__(input_key, bboxes_key, meta_keys,
                                        input_shape=input_shape, target_shape=target_shape,
                                        target_min=target_min, target_max=target_max)
        ClassNamesHandler.__init__(self, class_names=class_names)
        if background is not None:
            background = None if "background" not in self.class_indices else self.class_indices["background"]
        self.background = background
        self.color_map = color_map

    @property
    def background(self):
        """Returns the 'background' label value used in loss functions (can be ``None``)."""
        return self._background

    @background.setter
    def background(self, background):
        """Sets the 'background' label value for this segmentation task (can be ``None``)."""
        if background is not None:
            assert isinstance(background, int), "'background' value should be integer (index)"
            assert background not in self.class_indices.values() or self.class_indices["background"] == background, \
                "found 'background' value tied to another class label"
        self._background = background

    @property
    def color_map(self):
        """Returns the color map used to swap label indices for colors when displaying results."""
        return self._color_map

    @color_map.setter
    def color_map(self, color_map):
        """Sets the color map used to swap label indices for colors when displaying results."""
        if color_map is not None:
            assert isinstance(color_map, dict), "color map should be given as dictionary"
            self._color_map = {}
            assert all([isinstance(k, int) for k in color_map]) or all([isinstance(k, str) for k in color_map]), \
                "color map keys should be only class names or only class indices"
            for key, val in color_map.items():
                if isinstance(key, str):
                    if key == "background" and self.background is not None:
                        key = self.background
                    else:
                        assert key in self.class_indices, f"could not find color map key '{key}' in class names"
                        key = self.class_indices[key]
                assert key in self.class_indices.values() or key == self.background, f"unrecognized class index '{key}'"
                if isinstance(val, (list, tuple)):
                    val = np.asarray(val)
                assert isinstance(val, np.ndarray) and val.size == 3, "color values should be given as triplets"
                self._color_map[key] = val
            if self.background is not None and self.background not in self._color_map:
                self._color_map[self.background] = np.asarray([0, 0, 0])  # use black as default 'background' color
        else:
            self._color_map = {}

    def get_class_sizes(self, samples, bbox_format=None):
        """Given a list of samples, returns a map of element counts for each object type."""
        assert samples is not None and samples, "provided invalid sample list"
        elem_counts = {class_name: 0 for class_name in self.class_names}
        for sample_idx, sample in tqdm.tqdm(enumerate(samples), desc="cumulating bbox counts", total=len(samples)):
            if self.gt_key is None or self.gt_key not in sample:
                continue
            else:
                bboxes = sample[self.gt_key]
                if isinstance(bboxes, torch.Tensor):
                    bboxes = bboxes.cpu().numpy()
                if isinstance(bboxes, (np.ndarray, list, tuple)):
                    bboxes = [BoundingBox.decode(bbox, format=bbox_format)
                              if isinstance(bbox, (np.ndarray, list, tuple)) else bbox for bbox in bboxes]
                assert all([isinstance(bbox, BoundingBox) for bbox in bboxes]), "unrecognized sample bbox format"
                assert all([bbox.class_id in self.class_indices.values() for bbox in bboxes]), \
                    "bboxes contain unknown class ids"
                for cname, cval in self.class_indices.items():
                    elem_counts[cname] += len([b for b in bboxes if b.class_id == cval])
        return elem_counts

    def check_compat(self, task, exact=False):
        # type: (Detection, Optional[bool]) -> bool
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. If ``exact = True``, all fields will be checked for exact (perfect)
        compatibility (in this case, matching meta keys and class maps).
        """
        if isinstance(task, Detection):
            if not Regression.check_compat(self, task, exact=exact):
                return False
            return self.background == task.background and \
                all([cls in self.class_names for cls in task.class_names]) and \
                (not exact or (self.class_names == task.class_names and
                               self.color_map.keys() == task.color_map.keys() and
                               all([np.array_equal(self.color_map[k], task.color_map[k]) for k in self.color_map])))
        elif type(task) == Task:
            # if 'task' simply has no gt, compatibility rests on input key only
            return not exact and self.input_key == task.input_key and task.gt_key is None
        return False

    def get_compat(self, task):
        """Returns a task instance compatible with the current task and the given one."""
        assert isinstance(task, Detection) or type(task) == Task, \
            f"cannot create compatible task from types '{type(task)}' and '{type(self)}'"
        if isinstance(task, Detection):
            assert self.input_key == task.input_key, "input key mismatch, cannot create compatible task"
            assert self.gt_key is None or task.gt_key is None or self.gt_key == task.gt_key, \
                "gt key mismatch, cannot create compatible task"
            assert self.background == task.background, "background value mismatch, cannot create compatible task"
            meta_keys = list(set(self.meta_keys + task.meta_keys))
            # cannot use set for class names, order needs to stay intact!
            class_indices = {cname: cval for cname, cval in task.class_indices.items() if cname not in self.class_indices}
            class_indices = {**self.class_indices, **class_indices}
            color_map = {cname: cval for cname, cval in task.color_map.items() if cname not in self.color_map}
            color_map = {**self.color_map, **color_map}
            return Detection(class_names=class_indices, input_key=self.input_key, bboxes_key=self.gt_key, meta_keys=meta_keys,
                             input_shape=self.input_shape if self.input_shape is not None else task.input_shape,
                             target_shape=self.target_shape if self.target_shape is not None else task.target_shape,
                             target_min=self.target_min if self.target_min is not None else task.target_min,
                             target_max=self.target_max if self.target_max is not None else task.target_max,
                             background=self.background, color_map=color_map)
        elif type(task) == Task:
            assert self.check_compat(task), f"cannot create compatible task between:\n\t{str(self)}\n\t{str(task)}"
            meta_keys = list(set(self.meta_keys + task.meta_keys))
            return Detection(class_names=self.class_indices, input_key=self.input_key, bboxes_key=self.gt_key,
                             meta_keys=meta_keys, input_shape=self.input_shape, target_shape=self.target_shape,
                             target_min=self.target_min, target_max=self.target_max, background=self.background,
                             color_map=self.color_map)

    def __repr__(self):
        """Creates a print-friendly representation of a segmentation task."""
        color_map = {k: v.tolist() for k, v in self.color_map.items()}
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(class_names={repr(self.class_indices)}, input_key={repr(self.input_key)}, " + \
            f"bboxes_key={repr(self.gt_key)}, meta_keys={repr(self.meta_keys)}, " + \
            f"input_shape={repr(self.input_shape)}, target_shape={repr(self.target_shape)}, " + \
            f"target_min={repr(self.target_min)}, target_max={repr(self.target_max)}, " + \
            f"background={repr(self.background)}, color_map={repr(color_map)})"



import os

import mock

import thelper

CUR_DIR = os.path.abspath(os.path.dirname(__file__))
MOD_DIR = os.path.join(CUR_DIR, "mod")
MOD_INIT = os.path.join(MOD_DIR, "__init__.py")
PKG_DIR = os.path.join(MOD_DIR, "pkg")
PKG_INIT = os.path.join(PKG_DIR, "__init__.py")


def test_import_class_fully_qualified():
    c = thelper.nn.Module
    cn = "thelper.nn.Module"
    cls = thelper.utils.import_class(cn)
    assert cls is c


def test_import_class_path_qualified():
    assert not os.path.isfile(MOD_INIT), f"File {MOD_INIT} must not exist for this test evaluation"
    assert not os.path.isfile(PKG_INIT), f"File {PKG_INIT} must not exist for this test evaluation"
    cn = os.path.join(CUR_DIR, "mod.pkg.loader.RandomLoader")
    cls = thelper.utils.import_class(cn)
    assert cls is not None
    assert issubclass(cls, thelper.data.ImageFolderDataset)


# noinspection PyUnusedLocal
@mock.patch('thelper.utils.get_available_cuda_devices', return_value=[])    # no device available
@mock.patch('torch.load', return_value={'version': thelper.__version__})
def test_load_checkpoint_default_cpu_no_devices(mock_torch_load, mock_thelper_devices):
    checkpoint_file = 'dummy-checkpoint.pth'
    thelper.utils.load_checkpoint(checkpoint_file)
    mock_torch_load.assert_called_once_with(checkpoint_file, map_location='cpu')


# noinspection PyUnusedLocal
@mock.patch('thelper.utils.get_available_cuda_devices', return_value=[0])   # one device available
@mock.patch('torch.load', return_value={'version': thelper.__version__})
def test_load_checkpoint_not_default_cpu_with_devices(mock_torch_load, mock_thelper_devices):
    checkpoint_file = 'dummy-checkpoint.pth'
    thelper.utils.load_checkpoint(checkpoint_file)
    mock_torch_load.assert_called_once_with(checkpoint_file, map_location=None)



import numpy as np
import torch

import thelper


class LeNet(thelper.nn.Module):
    """LeNet CNN implementation.

    See http://yann.lecun.com/exdb/lenet/ for more information.

    This is NOT a modern architecture; it is only provided here for tutorial purposes.
    """

    def __init__(self, task, input_shape=(1, 28, 28), conv1_filters=6, conv2_filters=16,
                 hidden1_size=120, hidden2_size=84, output_size=10):
        # note: must always forward args to base class to keep backup
        super().__init__(task, **{k: v for k, v in vars().items() if k not in ["self", "task", "__class__"]})
        padding = 2 if input_shape[1] == 28 else 0
        self.baseline = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=input_shape[0], out_channels=conv1_filters, kernel_size=5, stride=1, padding=padding),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2),
            torch.nn.Conv2d(in_channels=conv1_filters, out_channels=conv2_filters, kernel_size=5, stride=1, padding=0),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2),
        )
        self.classifier_input_size = conv2_filters * 5 * 5
        self.classifier = torch.nn.Sequential(
            torch.nn.Linear(in_features=self.classifier_input_size, out_features=hidden1_size),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features=hidden1_size, out_features=hidden2_size),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features=hidden2_size, out_features=output_size),
        )

    def forward(self, x):
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x)
        feature_maps = self.baseline(x)
        embeddings = feature_maps.view(-1, self.classifier_input_size)
        logits = self.classifier(embeddings)
        return logits

    def set_task(self, task):
        assert isinstance(task, thelper.tasks.Classification), "missing impl for non-classif task type"
        num_classes = len(task.class_names)
        if num_classes != self.classifier[4].out_features:
            self.classifier = torch.nn.Sequential(
                torch.nn.Linear(in_features=self.classifier_input_size, out_features=self.classifier[0].out_features),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=self.classifier[2].in_features, out_features=self.classifier[2].out_features),
                torch.nn.ReLU(),
                torch.nn.Linear(in_features=self.classifier[4].in_features, out_features=num_classes),
            )
        self.task = task



"""Geospatial data parser & utilities module."""

import functools
import json
import logging
import math
import os
import pickle

import cv2 as cv
import gdal
import numpy as np
import ogr
import osr
import shapely
import tqdm

import thelper.data
import thelper.data.geo as geo

logger = logging.getLogger(__name__)


class VectorCropDataset(thelper.data.Dataset):
    """Abstract dataset used to combine geojson vector data and rasters."""

    def __init__(self, raster_path, vector_path, px_size=None, skew=None,
                 allow_outlying_vectors=True, clip_outlying_vectors=True,
                 vector_area_min=0.0, vector_area_max=float("inf"),
                 vector_target_prop=None, feature_buffer=None, master_roi=None,
                 srs_target="3857", raster_key="raster", mask_key="mask",
                 cleaner=None, cropper=None, force_parse=False,
                 reproj_rasters=False, reproj_all_cpus=True,
                 keep_rasters_open=True, transforms=None):
        # before anything else, create a hash to cache parsed data
        cache_hash = thelper.utils.get_params_hash(
            {k: v for k, v in vars().items() if not k.startswith("_") and k != "self"}) if not force_parse else None
        assert isinstance(raster_path, str), f"raster file/folder path should be given as string"
        assert isinstance(vector_path, str), f"vector file/folder path should be given as string"
        self.raster_path = raster_path
        self.vector_path = vector_path
        assert px_size is None or \
            (isinstance(px_size, (list, tuple)) and all([isinstance(i, (float, int)) for i in px_size]) and len(px_size) == 2) or \
            isinstance(px_size, (float, int)), "pixel size (resolution) must be float/int or list/tuple"
        self.px_size = (1.0, 1.0) if px_size is None else (float(px_size[0]), float(px_size[1])) \
            if isinstance(px_size, (list, tuple)) else (float(px_size), float(px_size))
        assert skew is None or \
            (isinstance(skew, (list, tuple)) and all([isinstance(i, (float, int)) for i in skew]) and len(skew) == 2) or \
            isinstance(skew, (float, int)), "pixel skew must be float/int or list/tuple"
        self.skew = (0.0, 0.0) if skew is None else (float(skew[0]), float(skew[1])) \
            if isinstance(skew, (list, tuple)) else (float(skew), float(skew))
        assert isinstance(allow_outlying_vectors, bool), "unexpected flag type"
        assert isinstance(clip_outlying_vectors, bool), "unexpected flag type"
        assert isinstance(force_parse, bool), "unexpected flag type"
        assert isinstance(reproj_rasters, bool), "unexpected flag type"
        assert isinstance(reproj_all_cpus, bool), "unexpected flag type"
        assert isinstance(keep_rasters_open, bool), "unexpected flag type"
        self.allow_outlying = allow_outlying_vectors
        self.clip_outlying = clip_outlying_vectors
        self.force_parse = force_parse
        self.reproj_rasters = reproj_rasters
        self.reproj_all_cpus = reproj_all_cpus
        self.keep_rasters_open = keep_rasters_open
        assert isinstance(vector_area_min, (float, int)) and vector_area_min >= 0, \
            "min surface filter value must be > 0"
        assert isinstance(vector_area_max, (float, int)) and vector_area_max >= vector_area_min, \
            "max surface filter value must be greater than minimum surface value"
        self.area_min = float(vector_area_min)
        self.area_max = float(vector_area_max)
        assert vector_target_prop is None or isinstance(vector_target_prop, dict), \
            "feature target props should be specified as dictionary of property name-value pairs for search"
        self.target_prop = {} if vector_target_prop is None else vector_target_prop
        assert feature_buffer is None or (isinstance(feature_buffer, (int, float)) and feature_buffer > 0), \
            "feature roi 'buffer' value should be strictly positive int/float"
        self.feature_buffer = feature_buffer
        assert isinstance(master_roi, (str, shapely.geometry.polygon.Polygon,
                                       shapely.geometry.multipolygon.MultiPolygon)) or master_roi is None, \
            "invalid master roi (should be path to geojson/shapefile or polygon object)"
        assert isinstance(srs_target, (str, int, osr.SpatialReference)), \
            "target EPSG SRS must be given as int/str"
        self.srs_target = srs_target
        if isinstance(self.srs_target, (str, int)):
            if isinstance(self.srs_target, str):
                self.srs_target = int(self.srs_target.replace("EPSG:", ""))
            srs_target_obj = osr.SpatialReference()
            srs_target_obj.ImportFromEPSG(self.srs_target)
            self.srs_target = srs_target_obj
        self.master_roi = geo.utils.parse_roi(master_roi, srs_target=self.srs_target) \
            if isinstance(master_roi, str) else master_roi
        assert isinstance(raster_key, str), "raster key must be given as string"
        self.raster_key = raster_key
        assert isinstance(mask_key, str), "mask key must be given as string"
        self.mask_key = mask_key
        super().__init__(transforms=transforms)
        self.rasters_data, self.coverage = self._parse_rasters(self.raster_path, self.srs_target, reproj_rasters)
        if self.master_roi is not None:
            self.coverage = self.coverage.intersection(self.master_roi)
        if cleaner is None:
            cleaner = functools.partial(self._default_feature_cleaner, area_min=self.area_min,
                                        area_max=self.area_max, target_prop=self.target_prop)
        self.features = self._parse_features(self.vector_path, self.srs_target, self.coverage, cache_hash,
                                             self.allow_outlying, self.clip_outlying, cleaner)
        if cropper is None:
            cropper = functools.partial(self._default_feature_cropper, px_size=self.px_size,
                                        skew=self.skew, feature_buffer=self.feature_buffer)
        self.samples = self._parse_crops(cropper, self.vector_path, cache_hash)
        # all keys already in sample dicts should be 'meta'; mask & raster will be added later
        meta_keys = list(set([k for s in self.samples for k in s]))
        if self.mask_key not in meta_keys:
            meta_keys.append(self.mask_key)
        # create default task without gt specification (this is a pretty basic parser)
        self.task = thelper.tasks.Task(input_key=self.raster_key, meta_keys=meta_keys)
        self.display_debug = False  # for internal debugging purposes only

    @staticmethod
    def _default_feature_cleaner(features, area_min, area_max, target_prop=None):
        """Flags geometric features as 'clean' based on some criteria (may be modified in derived classes)."""
        # note: we use a flag here instead of removing bad features so that end-users can still use them if needed
        for feature in tqdm.tqdm(features, desc="cleaning up features"):
            assert isinstance(feature, dict) and "clean" not in feature
            feature["clean"] = True
            if target_prop is not None and "properties" in feature and isinstance(feature["properties"], dict):
                if not all([k in feature["properties"] and feature["properties"][k] == v for k, v in target_prop.items()]):
                    feature["clean"] = False
            if not (area_min <= feature["geometry"].area <= area_max):
                feature["clean"] = False
        return features

    @staticmethod
    def _default_feature_cropper(features, rasters_data, coverage, srs_target, px_size, skew, feature_buffer):
        """Returns the samples for a set of features (may be modified in derived classes)."""
        # note: default behavior = just center on the feature, and pad if required by user
        samples = []
        clean_feats = [f for f in features if f["clean"]]
        srs_target_wkt = srs_target.ExportToWkt()
        for feature in tqdm.tqdm(clean_feats, desc="validating crop candidates"):
            assert feature["clean"]  # should not get here with bad features
            roi, roi_tl, roi_br, crop_width, crop_height = \
                geo.utils.get_feature_roi(feature["geometry"], px_size, skew, feature_buffer)
            # test all raster regions that touch the selected feature
            raster_hits = []
            for raster_idx, raster_data in enumerate(rasters_data):
                if raster_data["target_roi"].intersects(roi):
                    raster_hits.append(raster_idx)
            # make list of all other features that may be included in the roi
            roi_radius = np.linalg.norm(np.asarray(roi_tl) - np.asarray(roi_br)) / 2
            roi_features = [f for f in features if feature["centroid"].distance(f["centroid"]) <= roi_radius and
                            f["geometry"].intersects(roi)]
            # prepare actual 'sample' for crop generation at runtime
            samples.append({
                "features": roi_features,
                "focal": feature,
                "roi": roi,
                "roi_tl": roi_tl,
                "roi_br": roi_br,
                "raster_hits": raster_hits,
                "crop_width": crop_width,
                "crop_height": crop_height,
                "geotransform": np.asarray((roi_tl[0], px_size[0], skew[0],
                                            roi_tl[1], skew[1], px_size[1])),
                "srs": srs_target_wkt,
            })
        return samples

    @staticmethod
    def _parse_rasters(path, srs, reproj_rasters):
        """Parses rasters (geotiffs) and returns metadata/coverage information."""
        logger.info(f"parsing rasters from path '{path}'...")
        raster_paths = thelper.utils.get_file_paths(path, ".", allow_glob=True)
        rasters_data, coverage = geo.utils.parse_rasters(raster_paths, srs, reproj_rasters)
        assert rasters_data, f"could not find any usable rasters at '{raster_paths}'"
        logger.debug(f"rasters total coverage area = {coverage.area:.2f}")
        for idx, data in enumerate(rasters_data):
            logger.debug(f"raster #{idx + 1} area = {data['target_roi'].area:.2f}")
            # here, we enforce that raster datatypes/bandcounts match
            assert data["band_count"] == rasters_data[0]["band_count"], \
                f"parser expects that all raster band counts match" + \
                f"(found {str(data['band_count'])} and {str(rasters_data[0]['band_count'])})"
            assert data["data_type"] == rasters_data[0]["data_type"], \
                f"parser expects that all raster data types match" + \
                f"(found {str(data['data_type'])} and {str(rasters_data[0]['data_type'])})"
            data["to_target_transform"] = osr.CoordinateTransformation(data["srs"], srs)
            data["from_target_transform"] = osr.CoordinateTransformation(srs, data["srs"])
        return rasters_data, coverage

    @staticmethod
    def _parse_features(path, srs, roi, cache_hash, allow_outlying, clip_outlying, cleaner):
        """Parses vector files (geojsons) and returns geometry information."""
        logger.info(f"parsing vectors from path '{path}'...")
        assert os.path.isfile(path) and path.endswith("geojson"), \
            "vector file must be provided as geojson (shapefile support still incomplete)"
        cache_file_path = os.path.join(os.path.dirname(path), cache_hash + ".feats.pkl") \
            if cache_hash else None
        if cache_file_path is not None and os.path.exists(cache_file_path):
            logger.debug(f"parsing cached feature data from '{cache_file_path}'...")
            with open(cache_file_path, "rb") as fd:
                features = pickle.load(fd)
        else:
            with open(path) as vector_fd:
                vector_data = json.load(vector_fd)
            features = geo.utils.parse_geojson(vector_data, srs_target=srs, roi=roi,
                                               allow_outlying=allow_outlying, clip_outlying=clip_outlying)
            features = cleaner(features)
            if cache_file_path is not None:
                logger.debug(f"caching clean data to '{cache_file_path}'...")
                with open(cache_file_path, "wb") as fd:
                    pickle.dump(features, fd)
        logger.debug(f"cleanup resulted in {len([f for f in features if f['clean']])} features of interest")
        return features

    def _parse_crops(self, cropper, cache_file_path, cache_hash):
        """Parses crops based on prior feature/raster data.

        Each 'crop' corresponds to a sample that can be loaded at runtime.
        """
        logger.info(f"preparing crops...")
        cache_file_path = os.path.join(os.path.dirname(cache_file_path), cache_hash + ".crops.pkl") \
            if cache_hash else None
        if cache_file_path is not None and os.path.exists(cache_file_path):
            logger.debug(f"parsing cached crop data from '{cache_file_path}'...")
            with open(cache_file_path, "rb") as fd:
                samples = pickle.load(fd)
        else:
            samples = cropper(self.features, self.rasters_data, self.coverage, self.srs_target)
            if cache_file_path is not None:
                logger.debug(f"caching crop data to '{cache_file_path}'...")
                with open(cache_file_path, "wb") as fd:
                    pickle.dump(samples, fd)
        return samples

    def _process_crop(self, sample):
        """Returns a crop for a specific (internal) set of sampled features."""
        # remember: we assume that all rasters have the same intrinsic settings
        crop_datatype = geo.utils.GDAL2NUMPY_TYPE_CONV[self.rasters_data[0]["data_type"]]
        crop_size = (sample["crop_height"], sample["crop_width"], self.rasters_data[0]["band_count"])
        crop = np.ma.array(np.zeros(crop_size, dtype=crop_datatype), mask=np.ones(crop_size, dtype=np.uint8))
        mask = np.zeros(crop_size[0:2], dtype=np.uint8)
        crop_raster_gdal = gdal.GetDriverByName("MEM").Create("", crop_size[1], crop_size[0],
                                                              crop_size[2], self.rasters_data[0]["data_type"])
        crop_raster_gdal.SetGeoTransform(sample["geotransform"])
        crop_raster_gdal.SetProjection(self.srs_target.ExportToWkt())
        crop_mask_gdal = gdal.GetDriverByName("MEM").Create("", crop_size[1], crop_size[0], 1, gdal.GDT_Byte)
        crop_mask_gdal.SetGeoTransform(sample["geotransform"])
        crop_mask_gdal.SetProjection(self.srs_target.ExportToWkt())
        crop_mask_gdal.GetRasterBand(1).WriteArray(np.zeros(crop_size[0:2], dtype=np.uint8))
        ogr_dataset = ogr.GetDriverByName("Memory").CreateDataSource("mask")
        ogr_layer = ogr_dataset.CreateLayer("feature_mask", srs=self.srs_target)
        for feature in sample["features"]:
            ogr_feature = ogr.Feature(ogr_layer.GetLayerDefn())
            ogr_geometry = ogr.CreateGeometryFromWkt(feature["geometry"].wkt)
            ogr_feature.SetGeometry(ogr_geometry)
            ogr_layer.CreateFeature(ogr_feature)
        gdal.RasterizeLayer(crop_mask_gdal, [1], ogr_layer, burn_values=[1], options=["ALL_TOUCHED=TRUE"])
        np.copyto(dst=mask, src=crop_mask_gdal.GetRasterBand(1).ReadAsArray())
        for raster_idx in sample["raster_hits"]:
            rasterfile = geo.utils.open_rasterfile(self.rasters_data[raster_idx],
                                                   keep_rasters_open=self.keep_rasters_open)
            assert rasterfile.RasterCount == crop_size[2], "unexpected raster count"
            # using all cpus should be ok since we probably cant parallelize this loader anyway (swig serialization issues)
            options = ["NUM_THREADS=ALL_CPUS"] if self.reproj_all_cpus else []
            geo.utils.reproject_crop(rasterfile, crop_raster_gdal, crop_size, crop_datatype, reproj_opt=options, fill_nodata=True)
            for raster_band_idx in range(crop_raster_gdal.RasterCount):
                curr_band = crop_raster_gdal.GetRasterBand(raster_band_idx + 1)
                curr_band_array = curr_band.ReadAsArray()
                flag_mask = curr_band_array != curr_band.GetNoDataValue()
                np.copyto(dst=crop.data[:, :, raster_band_idx], src=curr_band_array, where=flag_mask)
                np.bitwise_and(crop.mask[:, :, raster_band_idx], np.invert(flag_mask), out=crop.mask[:, :, raster_band_idx])
        # ogr_dataset = None # close local fd
        # noinspection PyUnusedLocal
        crop_raster_gdal = None  # close local fd
        # noinspection PyUnusedLocal
        crop_mask_gdal = None  # close local fd
        # noinspection PyUnusedLocal
        rasterfile = None  # close input fd
        return crop, mask

    def __getitem__(self, idx):
        """Returns the data sample (a dictionary) for a specific (0-based) index."""
        if isinstance(idx, slice):
            return self._getitems(idx)
        assert idx < len(self.samples), "sample index is out-of-range"
        if idx < 0:
            idx = len(self.samples) + idx
        sample = self.samples[idx]
        crop, mask = self._process_crop(sample)
        if self.display_debug:
            crop = cv.cvtColor(crop, cv.COLOR_GRAY2BGR)
            mask = cv.cvtColor(mask, cv.COLOR_GRAY2BGR)
            mask[:, :, 1:3] = 0
            crop = cv.normalize(crop, dst=crop, alpha=0, beta=255, norm_type=cv.NORM_MINMAX, dtype=cv.CV_8U)
            mask = cv.normalize(mask, dst=mask, alpha=0, beta=255, norm_type=cv.NORM_MINMAX, dtype=cv.CV_8U)
            crop = np.bitwise_or(crop, mask)
        sample = {
            self.raster_key: np.array(crop.data, copy=True),
            self.mask_key: mask,
            **sample
        }
        if self.transforms:
            sample = self.transforms(sample)
        return sample


class TileDataset(VectorCropDataset):
    """Abstract dataset used to systematically tile vector data and rasters."""

    def __init__(self, raster_path, vector_path, tile_size, tile_overlap=0,
                 skip_empty_tiles=False, skip_nodata_tiles=True, px_size=None,
                 allow_outlying_vectors=True, clip_outlying_vectors=True,
                 vector_area_min=0.0, vector_area_max=float("inf"),
                 vector_target_prop=None, master_roi=None, srs_target="3857",
                 raster_key="raster", mask_key="mask", cleaner=None,
                 force_parse=False, reproj_rasters=False,
                 reproj_all_cpus=True, keep_rasters_open=True, transforms=None):
        # note1: input 'tile_size' must be given in pixels
        # note2: input 'tile_overlap' must be given in pixels
        # note3: input 'px_size' must be given in meters/degrees
        if isinstance(tile_size, (float, int)):
            tile_size = (tile_size, tile_size)
        assert isinstance(tile_size, (tuple, list)) and len(tile_size) == 2, \
            "invalid tile size (should be scalar or two-elem tuple)"
        assert all([t > 0 for t in tile_size]), "unexpected tile size value (should be positive)"
        tile_size = [float(t) for t in tile_size]  # convert all vals if necessary
        assert isinstance(tile_overlap, (float, int)) and tile_overlap >= 0, \
            "unexpected tile overlap (should be non-negative scalar)"
        tile_overlap = float(tile_overlap)
        assert isinstance(skip_empty_tiles, bool), "unexpected flag type (should be bool)"
        assert isinstance(skip_nodata_tiles, bool), "unexpected flag type (should be bool)"
        cropper = functools.partial(self._tile_cropper, tile_size=tile_size, tile_overlap=tile_overlap,
                                    skip_empty_tiles=skip_empty_tiles, skip_nodata_tiles=skip_nodata_tiles,
                                    keep_rasters_open=keep_rasters_open, px_size=px_size)
        super().__init__(raster_path=raster_path, vector_path=vector_path, px_size=px_size, skew=None,
                         allow_outlying_vectors=allow_outlying_vectors, clip_outlying_vectors=clip_outlying_vectors,
                         vector_area_min=vector_area_min, vector_area_max=vector_area_max, vector_target_prop=vector_target_prop,
                         master_roi=master_roi, srs_target=srs_target, raster_key=raster_key, mask_key=mask_key,
                         cleaner=cleaner, cropper=cropper, force_parse=force_parse, reproj_rasters=reproj_rasters,
                         reproj_all_cpus=reproj_all_cpus, keep_rasters_open=keep_rasters_open, transforms=transforms)

    @staticmethod
    def _tile_cropper(features, rasters_data, coverage, srs_target, tile_size, tile_overlap,
                      skip_empty_tiles, skip_nodata_tiles, keep_rasters_open, px_size):
        """Returns the ROI information for a given feature (may be modified in derived classes)."""
        # instead of iterating over features to generate samples, we tile the raster(s)
        # note: the 'coverage' geometry should already be in the target srs
        roi_tl, roi_br = geo.utils.get_feature_bbox(coverage)
        roi_geotransform = (roi_tl[0], px_size[0], 0.0,
                            roi_tl[1], 0.0, px_size[1])
        srs_target_wkt = srs_target.ExportToWkt()
        # remember: we assume that all rasters have the same intrinsic settings
        crop_datatype = geo.utils.GDAL2NUMPY_TYPE_CONV[rasters_data[0]["data_type"]]
        crop_raster_gdal = gdal.GetDriverByName("MEM").Create("",
                                                              int(round(tile_size[1])),
                                                              int(round(tile_size[0])),
                                                              rasters_data[0]["band_count"],
                                                              rasters_data[0]["data_type"])
        crop_raster_gdal.SetProjection(srs_target_wkt)
        samples = []
        crop_id = 0
        roi_px_br = geo.utils.get_pxcoord(roi_geotransform, *roi_br)
        nb_iter_y = int(math.ceil((roi_px_br[1] + tile_overlap) / (tile_size[1] - tile_overlap)))
        nb_iter_x = int(math.ceil((roi_px_br[0] + tile_overlap) / (tile_size[0] - tile_overlap)))
        pbar = tqdm.tqdm(total=nb_iter_y * nb_iter_x, desc="validating crop candidates")
        roi_offset_px_y = -tile_overlap
        while roi_offset_px_y < roi_px_br[1]:
            roi_offset_px_x = -tile_overlap
            while roi_offset_px_x < roi_px_br[0]:
                pbar.update(1)
                crop_px_tl = (roi_offset_px_x, roi_offset_px_y)
                crop_px_br = (crop_px_tl[0] + tile_size[0], crop_px_tl[1] + tile_size[1])
                crop_tl = geo.utils.get_geocoord(roi_geotransform, *crop_px_tl)
                crop_br = geo.utils.get_geocoord(roi_geotransform, *crop_px_br)
                crop_geom = shapely.geometry.Polygon([crop_tl, (crop_br[0], crop_tl[1]),
                                                      crop_br, (crop_tl[0], crop_br[1])])
                crop_geotransform = (crop_tl[0], px_size[0], 0.0,
                                     crop_tl[1], 0.0, px_size[1])
                crop_raster_gdal.SetGeoTransform(crop_geotransform)
                raster_hits = []
                found_valid_intersection = False or not skip_nodata_tiles
                for raster_idx, raster_data in enumerate(rasters_data):
                    if raster_data["target_roi"].intersects(crop_geom):
                        if not found_valid_intersection:
                            rasterfile = geo.utils.open_rasterfile(raster_data, keep_rasters_open=keep_rasters_open)
                            # yeah, we reproject the crop, preprocessing is slow, deal with it
                            geo.utils.reproject_crop(rasterfile, crop_raster_gdal, tile_size, crop_datatype, fill_nodata=True)
                            for raster_band_idx in range(crop_raster_gdal.RasterCount):
                                curr_band = crop_raster_gdal.GetRasterBand(raster_band_idx + 1)
                                found_valid_intersection = found_valid_intersection or \
                                    np.count_nonzero(curr_band.ReadAsArray() != curr_band.GetNoDataValue()) > 0
                        raster_hits.append(raster_idx)
                if raster_hits and found_valid_intersection:
                    crop_centroid = crop_geom.centroid
                    crop_radius = np.linalg.norm(np.asarray(crop_tl) - np.asarray(crop_br)) / 2
                    crop_features = []
                    for f in features:
                        if f["geometry"].distance(crop_centroid) > crop_radius:
                            continue
                        inters = f["geometry"].intersection(crop_geom)
                        if inters.is_empty:
                            continue
                        crop_features.append(f)
                    if crop_features or not skip_empty_tiles:
                        # prepare actual 'sample' for crop generation at runtime
                        samples.append({
                            "features": crop_features,
                            "id": crop_id,
                            "roi": crop_geom,
                            "roi_tl": crop_tl,
                            "roi_br": crop_br,
                            "raster_hits": raster_hits,
                            "crop_width": int(round(tile_size[0])),
                            "crop_height": int(round(tile_size[1])),
                            "geotransform": np.asarray(crop_geotransform),
                        })
                crop_id += 1
                roi_offset_px_x += tile_size[0] - tile_overlap
            roi_offset_px_y += tile_size[1] - tile_overlap
        return samples



import copy
import math
import random

import numpy as np
import pytest
import torch

import thelper


class CustomSampler(torch.utils.data.sampler.RandomSampler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.epoch = None

    def set_epoch(self, epoch=0):
        self.epoch = epoch


class CustomTransformer:

    def __init__(self):
        self.epoch = None

    def __call__(self, x):
        return x

    def set_epoch(self, epoch=0):
        self.epoch = epoch


class CustomTensorDataset(torch.utils.data.Dataset):

    def __init__(self, tensor):
        self.tensor = tensor
        self.epoch = None
        self.transforms = CustomTransformer()

    def __getitem__(self, index):
        rand_max = 2 ** 16 - 1
        return self.transforms((self.tensor[index],
                                np.random.randint(rand_max),
                                torch.randint(0, rand_max, size=(1,)).item(),
                                random.randint(0, rand_max)))

    def __len__(self):
        return self.tensor.size(0)

    def set_epoch(self, epoch=0):
        self.epoch = epoch


@pytest.fixture
def tensor_dataset():
    return CustomTensorDataset(torch.Tensor([torch.Tensor([v]) for v in range(100)]))


@pytest.mark.parametrize("num_workers", [0, 1, 2])
def test_tensor_loader_interface(tensor_dataset, num_workers):
    loader = thelper.data.DataLoader(tensor_dataset, num_workers=num_workers, batch_size=2, seeds={
        "torch": 0, "numpy": 0, "random": 0
    })
    assert len(loader) == len(tensor_dataset) // 2
    loader = thelper.data.DataLoader(tensor_dataset, sampler=CustomSampler(tensor_dataset), num_workers=num_workers, batch_size=2, seeds={
        "torch": 0, "numpy": 0, "random": 0
    })
    assert len(loader) == len(tensor_dataset) // 2
    assert loader.sample_count == len(tensor_dataset)
    with pytest.raises(AssertionError):
        _ = thelper.data.DataLoader(tensor_dataset, seeds=0)
    with pytest.raises(AssertionError):
        _ = thelper.data.DataLoader(tensor_dataset, epoch=None)
    rand_vals = []
    assert loader.epoch == 0
    for batch_idx, batch in enumerate(loader):
        if batch_idx == 0:
            rand_vals = batch[1:4]
        else:
            # pretty unlikely to get triple collision...
            assert not all([torch.all(torch.eq(rand_vals[idx], batch[1:4][idx])) for idx in range(3)])
        assert loader.sampler.epoch == 0
        assert loader.dataset.epoch == 0
        assert loader.dataset.transforms.epoch == 0
        assert batch[0].shape == (2,)
    assert loader.epoch == 1
    with pytest.raises(AssertionError):
        loader.set_epoch(None)
    loader.set_epoch(0)
    assert loader.epoch == 0
    for batch_idx, batch in enumerate(loader):
        if batch_idx == 0:
            assert all([torch.all(torch.eq(rand_vals[idx], batch[1:4][idx])) for idx in range(3)])
        assert batch[0].shape == (2,)
    assert loader.epoch == 1


class DummyClassifDataset(thelper.data.Dataset):
    def __init__(self, nb_samples, nb_classes, subset, transforms=None, deepcopy=False):
        super().__init__(transforms=transforms, deepcopy=deepcopy)
        inputs = torch.randint(0, 2 ** 16 - 1, size=(nb_samples, 1))
        labels = torch.remainder(torch.randperm(nb_samples), nb_classes)
        self.samples = [{"input": inputs[idx], "label": labels[idx], "idx": idx, "subset": subset} for idx in range(nb_samples)]
        self.task = thelper.tasks.Classification([str(idx) for idx in range(nb_classes)], "input", "label", meta_keys=["idx", "subset"])

    def __getitem__(self, idx):
        return self.samples[idx]


@pytest.fixture
def class_split_config():
    return {
        "datasets": {
            "dataset_A": DummyClassifDataset(1000, 10, "A"),
            "dataset_B": DummyClassifDataset(1000, 10, "B"),
            "dataset_C": DummyClassifDataset(1000, 10, "C")
        },
        "loaders": {
            "batch_size": 32,
            "train_split": {
                "dataset_A": 0.5,
                "dataset_B": 0.7
            },
            "valid_split": {
                "dataset_A": 0.4,
                "dataset_B": 0.3
            },
            "test_split": {
                "dataset_A": 0.1,
                "dataset_C": 1.0
            }
        }
    }


def test_classif_split(class_split_config):
    task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(class_split_config)
    assert task.check_compat(class_split_config["datasets"]["dataset_A"].task, exact=True)
    assert task.check_compat(class_split_config["datasets"]["dataset_B"].task, exact=True)
    assert task.check_compat(class_split_config["datasets"]["dataset_C"].task, exact=True)
    train_samples, valid_samples, test_samples = {}, {}, {}
    train_class_counts, valid_class_counts, test_class_counts = [0] * 10, [0] * 10, [0] * 10
    for loader, samples, class_counts in [(train_loader, train_samples, train_class_counts),
                                          (valid_loader, valid_samples, valid_class_counts),
                                          (test_loader, test_samples, test_class_counts)]:
        for batch in loader:
            for idx in range(batch["input"].size(0)):
                name = batch["subset"][idx] + str(batch["idx"][idx].item())
                assert name not in samples
                label = batch["label"][idx].item()
                class_counts[label] += 1
                samples[name] = label
    assert not bool(set(train_samples) & set(valid_samples))
    assert not bool(set(train_samples) & set(test_samples))
    assert not bool(set(valid_samples) & set(test_samples))
    assert all([count == train_class_counts[0] for count in train_class_counts])
    assert all([count == valid_class_counts[0] for count in valid_class_counts])
    assert all([count == test_class_counts[0] for count in test_class_counts])


def test_classif_split_no_balancing(class_split_config):
    class_split_config["loaders"]["skip_class_balancing"] = True
    task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(class_split_config)
    assert task.check_compat(class_split_config["datasets"]["dataset_A"].task, exact=True)
    assert task.check_compat(class_split_config["datasets"]["dataset_B"].task, exact=True)
    assert task.check_compat(class_split_config["datasets"]["dataset_C"].task, exact=True)
    train_samples, valid_samples, test_samples = {}, {}, {}
    for loader, samples in [(train_loader, train_samples), (valid_loader, valid_samples), (test_loader, test_samples)]:
        for batch in loader:
            for idx in range(batch["input"].size(0)):
                name = batch["subset"][idx] + str(batch["idx"][idx].item())
                assert name not in samples
                label = batch["label"][idx].item()
                samples[name] = label
    assert not bool(set(train_samples) & set(valid_samples))
    assert not bool(set(train_samples) & set(test_samples))
    assert not bool(set(valid_samples) & set(test_samples))


def collate_fn(*args, **kwargs):  # pragma: no cover
    return torch.utils.data.dataloader.default_collate(*args, **kwargs)


def test_loader_factory(class_split_config, mocker):
    fake_loaders_config = copy.deepcopy(class_split_config["loaders"])
    fake_loaders_config["train_batch_size"] = 16
    with pytest.raises(AssertionError):
        _ = thelper.data.loaders.LoaderFactory(fake_loaders_config)
    factory = thelper.data.loaders.LoaderFactory(class_split_config["loaders"])
    assert factory.train_split["dataset_A"] == 0.5 and factory.valid_split["dataset_A"] == 0.4 and factory.test_split["dataset_A"] == 0.1
    assert factory.train_split["dataset_B"] == 0.7 and factory.valid_split["dataset_B"] == 0.3
    assert factory.test_split["dataset_C"] == 1.0
    loaders_config_drop = copy.deepcopy(class_split_config["loaders"])
    loaders_config_drop["drop_last"] = True
    factory = thelper.data.loaders.LoaderFactory(loaders_config_drop)
    assert factory.drop_last
    loaders_config_sampler = copy.deepcopy(class_split_config["loaders"])
    loaders_config_sampler["sampler"] = {
        "type": "torch.utils.data.sampler.RandomSampler"
    }
    factory = thelper.data.loaders.LoaderFactory(loaders_config_sampler)
    assert factory.sampler_type == torch.utils.data.sampler.RandomSampler
    assert factory.train_sampler
    _ = mocker.patch("thelper.transforms.load_transforms", return_value="dummy")
    loaders_config_transfs = copy.deepcopy(class_split_config["loaders"])
    loaders_config_transfs["base_transforms"] = [
        {
            "operation": "thelper.transforms.RandomResizedCrop",
            "params": {
                "output_size": [100, 100]
            }
        },
        {
            "operation": "thelper.transforms.CenterCrop",
            "params": {
                "size": [200, 200]
            }
        }
    ]
    factory = thelper.data.loaders.LoaderFactory(loaders_config_transfs)
    assert factory.base_transforms is not None
    assert factory.get_base_transforms() == factory.base_transforms
    loaders_config_transfs["train_augments"] = {"append": True, "transforms": [
        {
            "operation": "thelper.transforms.RandomResizedCrop",
            "params": {
                "output_size": [100, 100]
            }
        },
        {
            "operation": "thelper.transforms.CenterCrop",
            "params": {
                "size": [200, 200]
            }
        }
    ]}
    factory = thelper.data.loaders.LoaderFactory(loaders_config_transfs)
    assert factory.train_augments is not None and factory.train_augments_append
    fake_loaders_split_config = copy.deepcopy(class_split_config["loaders"])
    del fake_loaders_split_config["train_split"]
    del fake_loaders_split_config["valid_split"]
    del fake_loaders_split_config["test_split"]
    with pytest.raises(AssertionError):
        _ = thelper.data.loaders.LoaderFactory(fake_loaders_split_config)
    normalize_query = mocker.patch("thelper.utils.query_yes_no", return_value=True)
    denormalized_loaders_config = copy.deepcopy(class_split_config["loaders"])
    denormalized_loaders_config["train_split"]["dataset_A"] = 1.5
    factory = thelper.data.loaders.LoaderFactory(denormalized_loaders_config)
    assert math.isclose(factory.train_split["dataset_A"] + factory.valid_split["dataset_A"] + factory.test_split["dataset_A"], 1)
    denormalized_loaders_config["train_split"]["dataset_A"] = 0.1
    factory = thelper.data.loaders.LoaderFactory(denormalized_loaders_config)
    assert math.isclose(factory.train_split["dataset_A"] + factory.valid_split["dataset_A"] + factory.test_split["dataset_A"], 1)
    assert normalize_query.call_count == 1
    loaders_collate_config = copy.deepcopy(class_split_config["loaders"])
    loaders_collate_config["collate_fn"] = 0
    with pytest.raises(AssertionError):
        _ = thelper.data.loaders.LoaderFactory(loaders_collate_config)
    loaders_collate_config["collate_fn"] = collate_fn
    factory = thelper.data.loaders.LoaderFactory(loaders_collate_config)
    assert factory.train_collate_fn == collate_fn
    loaders_collate_config["collate_fn"] = {
        "type": "torch.utils.data.dataloader.default_collate"
    }
    factory = thelper.data.loaders.LoaderFactory(loaders_collate_config)
    assert factory.train_collate_fn == torch.utils.data.dataloader.default_collate
    loaders_collate_config["collate_fn"] = "torch.utils.data.dataloader.default_collate"
    factory = thelper.data.loaders.LoaderFactory(loaders_collate_config)
    assert factory.train_collate_fn == torch.utils.data.dataloader.default_collate
    loaders_collate_config["train_collate_fn"] = "torch.utils.data.dataloader.default_collate"
    with pytest.raises(AssertionError):
        _ = thelper.data.loaders.LoaderFactory(loaders_collate_config)
    loaders_seed_config = copy.deepcopy(class_split_config["loaders"])
    loaders_seed_config["test_seed"] = 0
    loaders_seed_config["valid_seed"] = 1
    loaders_seed_config["torch_seed"] = 2
    loaders_seed_config["numpy_seed"] = 3
    loaders_seed_config["random_seed"] = "4"
    with pytest.raises(AssertionError):
        _ = thelper.data.loaders.LoaderFactory(loaders_seed_config)
    loaders_seed_config["random_seed"] = 4
    factory = thelper.data.loaders.LoaderFactory(loaders_seed_config)
    assert factory.seeds["test"] == 0
    assert factory.seeds["valid"] == 1
    assert factory.seeds["torch"] == 2
    assert factory.seeds["numpy"] == 3
    assert factory.seeds["random"] == 4



"""Transformation operations module.

All transforms should aim to be compatible with both numpy arrays and
PyTorch tensors. By default, images are processed using ``__call__``,
meaning that for a given transformation ``t``, we apply it via::

    image_transformed = t(image)

All important parameters for an operation should also be passed in the
constructor and exposed in the operation's ``__repr__`` function so that
external parsers can discover exactly how to reproduce their behavior. For
now, these representations are used for debugging more than anything else.
"""

import copy
import itertools
import logging
import math

import cv2 as cv
import numpy as np
import PIL.Image
import torch
import torchvision.transforms.functional
import torchvision.utils

import thelper.utils

logger = logging.getLogger(__name__)


class NoTransform:
    """Used to flag some ops that should not be externally wrapped for sample/key handling."""

    def __call__(self, sample):
        """Identity transform."""
        return sample

    def invert(self, sample):
        """Identity transform."""
        return sample

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + "()"


class ToNumpy:
    """Converts and returns an image in numpy format from a ``torch.Tensor`` or ``PIL.Image`` format.

    This operation is deterministic. The returns image will always be encoded as HxWxC, where
    if the input has three channels, the ordering might be optionally changed.

    Attributes:
        reorder_bgr: specifies whether the channels should be reordered in OpenCV format.
    """

    def __init__(self, reorder_bgr=False):
        """Initializes transformation parameters."""
        self.reorder_bgr = reorder_bgr

    def __call__(self, sample):
        """Converts and returns an image in numpy format.

        Args:
            sample: the image to convert; should be a tensor, numpy array, or PIL image.

        Returns:
            The numpy-converted image.
        """
        assert isinstance(sample, (np.ndarray, torch.Tensor, PIL.Image.Image)), \
            "unknown image type, cannot process sample"
        if isinstance(sample, np.ndarray):
            pass  # no transform needed, channel reordering done at end
        elif isinstance(sample, torch.Tensor):
            sample = sample.cpu().numpy()
        elif isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        return sample[..., ::-1] if self.reorder_bgr else sample

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as the original data type is unknown."""
        raise RuntimeError("cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(reorder_bgr={self.reorder_bgr})"


class CenterCrop:
    """Returns a center crop from a given image via OpenCV and numpy.

    This operation is deterministic. The code relies on OpenCV, meaning the border arguments
    must be compatible with ``cv2.copyMakeBorder``.

    Attributes:
        size: the size of the target crop (tuple of width, height).
        relative: specifies whether the target crop size is relative to the image size or not.
        bordertype: argument forwarded to ``cv2.copyMakeBorder``.
        borderval: argument forwarded to ``cv2.copyMakeBorder``.
    """

    def __init__(self, size, bordertype=cv.BORDER_CONSTANT, borderval=0):
        """Validates and initializes center crop parameters.

        Args:
            size: size of the target crop, provided as tuple or list. If integer values are used,
                the size is assumed to be absolute. If floating point values are used, the size is
                assumed to be relative, and will be determined dynamically for each sample. If a
                tuple is used, it is assumed to be (width, height).
            bordertype: border copy type to use when the image is too small for the required crop size.
                See ``cv2.copyMakeBorder`` for more information.
            borderval: border value to use when the image is too small for the required crop size. See
                ``cv2.copyMakeBorder`` for more information.
        """
        assert isinstance(size, (tuple, list, int, float)), \
            "unexpected center crop dim input type (need tuple or int/float)"
        if isinstance(size, (tuple, list)):
            assert len(size) == 2, "expected center crop dim input as 2-item tuple (width,height)"
            self.size = tuple(size)
        elif isinstance(size, (int, float)):
            self.size = (size, size)
        assert self.size[0] > 0 and self.size[1] > 0, "crop dimensions must be strictly positive"
        assert isinstance(self.size[0], (float, int)), \
            "unexpected center crop dim input type (need tuple or int/float)"
        self.relative = isinstance(self.size[0], float)
        self.bordertype = thelper.utils.import_class(bordertype) if isinstance(bordertype, str) else bordertype
        self.borderval = borderval

    def __call__(self, sample):
        """Extracts and returns a central crop from the provided image.

        Args:
            sample: the image to generate the crop from; should be a 2d or 3d numpy array.

        Returns:
            The center crop.
        """
        assert isinstance(sample, np.ndarray), f"sample type should be np.ndarray (got {type(sample)})"
        assert 2 <= sample.ndim <= 3, "bad input dimensions; must be 2-d, or 3-d (with channels)"
        crop_height = int(round(self.size[1] * sample.shape[0])) if self.relative else self.size[1]
        crop_width = int(round(self.size[0] * sample.shape[1])) if self.relative else self.size[0]
        tl = [sample.shape[1] // 2 - crop_width // 2, sample.shape[0] // 2 - crop_height // 2]
        br = [tl[0] + crop_width, tl[1] + crop_height]
        return thelper.draw.safe_crop(sample, tl, br, self.bordertype, self.borderval)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as data loss is incurred during image transformation."""
        raise RuntimeError("cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(size={self.size}, bordertype={self.bordertype}, borderval={self.borderval})"


class RandomResizedCrop:
    """Returns a resized crop of a randomly selected image region.

    This operation is stochastic, and thus cannot be inverted. Each time the operation is called,
    a random check will determine whether a transformation is applied or not. The code relies on
    OpenCV, meaning the interpolation arguments must be compatible with ``cv2.resize``.

    Attributes:
        output_size: size of the output crop, provided as a single element (``edge_size``) or as a
            two-element tuple or list (``[width, height]``). If integer values are used, the size is
            assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
            size is assumed to be relative to the original image size, and will be determined at
            execution time for each sample. If set to ``None``, the crop will not be resized.
        input_size: range of the input region sizes, provided as a pair of elements
            (``[min_edge_size, max_edge_size]``) or as a pair of tuples or lists
            (``[[min_width, min_height], [max_width, max_height]]``). If the pair-of-pairs format is
            used, the ``ratio`` argument cannot be used. If integer values are used, the ranges are
            assumed to be absolute. If floating point values are used (i.e. in [0,1]), the ranges
            are assumed to be relative to the original image size, and will be determined at
            execution time for each sample.
        ratio: range of minimum/maximum input region aspect ratios to use. This argument cannot be
            used if the pair-of-pairs format is used for the ``input_size`` argument.
        probability: the probability that the transformation will be applied when called; if not
            applied, the returned image will be the original.
        random_attempts: the number of random sampling attempts to try before reverting to center
            or most-probably-valid crop generation.
        min_roi_iou: minimum roi intersection over union (IoU) required for accepting a tile (in [0,1]).
        flags: interpolation flag forwarded to ``cv2.resize``.
    """

    def __init__(self, output_size, input_size=(0.08, 1.0), ratio=(0.75, 1.33), probability=1.0,
                 random_attempts=10, min_roi_iou=1.0, flags=cv.INTER_LINEAR):
        """Validates and initializes center crop parameters.

        Args:
            output_size: size of the output crop, provided as a single element (``edge_size``) or as a
                two-element tuple or list (``[width, height]``). If integer values are used, the size is
                assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
                size is assumed to be relative to the original image size, and will be determined at
                execution time for each sample. If set to ``None``, the crop will not be resized.
            input_size: range of the input region sizes, provided as a pair of elements
                (``[min_edge_size, max_edge_size]``) or as a pair of tuples or lists
                (``[[min_width, min_height], [max_width, max_height]]``). If the pair-of-pairs format is
                used, the ``ratio`` argument cannot be used. If integer values are used, the ranges are
                assumed to be absolute. If floating point values are used (i.e. in [0,1]), the ranges
                are assumed to be relative to the original image size, and will be determined at
                execution time for each sample.
            ratio: range of minimum/maximum input region aspect ratios to use. This argument cannot be
                used if the pair-of-pairs format is used for the ``input_size`` argument.
            probability: the probability that the transformation will be applied when called; if not
                applied, the returned image will be the original.
            random_attempts: the number of random sampling attempts to try before reverting to center
                or most-probably-valid crop generation.
            min_roi_iou: minimum roi intersection over union (IoU) required for producing a tile.
            flags: interpolation flag forwarded to ``cv2.resize``.
        """
        if output_size is None or isinstance(output_size, (tuple, list)):
            if output_size is not None:
                assert len(output_size) == 2, "expected output size to be two-element list or tuple, or single scalar"
                assert all([isinstance(s, int) for s in output_size]) or all([isinstance(s, float) for s in output_size]), \
                    "expected output size pair elements to be the same type (int or float)"
            self.output_size = output_size
        elif isinstance(output_size, (int, float)):
            self.output_size = (output_size, output_size)
        else:
            raise TypeError("unexpected output size type (need tuple/list/int/float)")
        if self.output_size is not None:
            for s in self.output_size:
                assert ((isinstance(s, float) and 0 < s <= 1) or (isinstance(s, int) and s > 0)), \
                    f"invalid output size value ({str(s)})"
        assert isinstance(input_size, (tuple, list)) and len(input_size) == 2, \
            "expected input size to be provided as a pair of elements or a pair of tuples/lists"
        if all([isinstance(s, int) for s in input_size]) or all([isinstance(s, float) for s in input_size]):
            if ratio is not None and isinstance(ratio, (tuple, list)):
                assert len(ratio) == 2, "invalid ratio tuple/list length (expected two elements)"
                assert all([isinstance(r, float) for r in ratio]), "expected ratio pair elements to be float values"
                self.ratio = (min(ratio), max(ratio))
            elif ratio is not None and isinstance(ratio, float):
                self.ratio = (ratio, ratio)
            else:
                raise TypeError("invalid aspect ratio, expected 2-element tuple/list or single float")
            self.input_size = ((min(input_size), min(input_size)), (max(input_size), max(input_size)))
        elif all([isinstance(s, tuple) and len(s) == 2 for s in input_size]) or \
                all([isinstance(s, list) and len(s) == 2 for s in input_size]):
            assert ratio is None or (isinstance(ratio, (tuple, list)) and len(ratio) == 0), \
                "cannot specify input sizes in two-element tuples/lists and also provide aspect ratios"
            for t in input_size:
                for s in t:
                    assert isinstance(s, (int, float)) and not isinstance(type(s), type(input_size[0][0])), \
                        "input sizes should all be same type, either int or float"
            self.input_size = ((min(input_size[0][0], input_size[1][0]), min(input_size[0][1], input_size[1][1])),
                               (max(input_size[0][0], input_size[1][0]), max(input_size[0][1], input_size[1][1])))
            self.ratio = None  # ignored since input_size contains all necessary info
        else:
            raise TypeError("expected input size to be two-elem list/tuple of int/float or two-element list/tuple of int/float")
        for t in self.input_size:
            assert ((isinstance(t, float) and 0 < s <= 1) or (isinstance(t, int) and s > 0)) or \
                all([((isinstance(s, float) and 0 < s <= 1) or (isinstance(s, int) and s > 0)) for s in t]), \
                "invalid input size value"
        assert 0 <= probability <= 1, "invalid probability value (should be in [0,1]"
        self.probability = probability
        assert random_attempts > 0, "invalid random_attempts value (should be > 0)"
        self.random_attempts = random_attempts
        assert isinstance(min_roi_iou, float) and 0 <= min_roi_iou <= 1, \
            "invalid minimum roi IoU score (should be float in [0,1])"
        self.min_roi_iou = min_roi_iou
        self.flags = thelper.utils.import_class(flags) if isinstance(flags, str) else flags
        self.warned_no_crop_found_with_roi = False

    def __call__(self, image, roi=None, mask=None, bboxes=None):
        """Extracts and returns a random (resized) crop from the provided image.

        Args:
            image: the image to generate the crop from. If given as a 2-element list, it is assumed to
                contain both the image and the roi (passed through a composer).
            roi: the roi to check tile intersections with (may be ``None``).
            mask: a mask to crop simultaneously with the input image (may be ``None``).
            bboxes: a list or array of bounding boxes to crop with the input image (may be ``None``).

        Returns:
            The randomly selected and resized crop. If mask and/or bboxes is given, the output will be
            a dictionary containing the results under the ``image``, ``mask``, and ``bboxes`` keys.
        """
        if isinstance(image, list) and len(image) == 2:
            assert roi is None, "roi provided twice"
            # we assume that the roi was given as the 2nd element of the list
            image, roi = image[0], image[1]
        assert isinstance(image, (PIL.Image.Image, np.ndarray)), \
            "image type should be np.ndarray or PIL image"
        if isinstance(image, PIL.Image.Image):
            image = np.asarray(image)
        if self.probability < 1 and np.random.uniform(0, 1) > self.probability:
            return image
        if bboxes is not None:
            raise NotImplementedError
        image_height, image_width = image.shape[0], image.shape[1]
        target_height, target_width = None, None
        target_row, target_col = None, None
        for attempt in range(self.random_attempts):
            if self.ratio is None:
                target_width = np.random.uniform(self.input_size[0][0], self.input_size[1][0])
                target_height = np.random.uniform(self.input_size[0][1], self.input_size[1][1])
                if isinstance(self.input_size[0][0], float):
                    target_width *= image_width
                    target_height *= image_height
                target_width = int(round(target_width))
                target_height = int(round(target_height))
            elif isinstance(self.input_size[0][0], (int, float)):
                if isinstance(self.input_size[0][0], float):
                    area = image_height * image_width
                    target_area = np.random.uniform(self.input_size[0][0], self.input_size[1][0]) * area
                else:
                    target_area = np.random.uniform(self.input_size[0][0], self.input_size[1][0]) ** 2
                aspect_ratio = np.random.uniform(*self.ratio)
                target_width = int(round(math.sqrt(target_area * aspect_ratio)))
                target_height = int(round(math.sqrt(target_area / aspect_ratio)))
                if np.random.random() < 0.5:
                    target_width, target_height = target_height, target_width
            else:
                raise RuntimeError("unhandled crop strategy")
            if target_width <= image_width and target_height <= image_height:
                target_col = np.random.randint(min(0, image_width - target_width), max(0, image_width - target_width) + 1)
                target_row = np.random.randint(min(0, image_height - target_height), max(0, image_height - target_height) + 1)
                if roi is None:
                    break
                roi = thelper.draw.safe_crop(roi, (target_col, target_row), (target_col + target_width, target_row + target_height))
                if np.count_nonzero(roi) >= target_width * target_height * self.min_roi_iou:
                    break
        if target_row is None or target_col is None:
            # fallback, use centered crop
            target_width = target_height = min(image.shape[0], image.shape[1])
            target_col = (image.shape[1] - target_width) // 2
            target_row = (image.shape[0] - target_height) // 2
            if roi is not None and not self.warned_no_crop_found_with_roi:
                logger.warning("random resized crop failing to find proper ROI matches after max attempt count")
                self.warned_no_crop_found_with_roi = True
        crop = thelper.draw.safe_crop(image, (target_col, target_row), (target_col + target_width, target_row + target_height))
        if self.output_size is None:
            return crop
        elif isinstance(self.output_size[0], float):
            output_width = int(round(self.output_size[0] * image.shape[1]))
            output_height = int(round(self.output_size[1] * image.shape[0]))
            return cv.resize(crop, (output_width, output_height), interpolation=self.flags)
        elif isinstance(self.output_size[0], int):
            return cv.resize(crop, (self.output_size[0], self.output_size[1]), interpolation=self.flags)
        else:
            raise RuntimeError("unhandled crop strategy")

    def invert(self, image):
        """Specifies that this operation cannot be inverted, as data loss is incurred during image transformation."""
        raise RuntimeError("cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(output_size={self.output_size}, input_size={self.input_size}, ratio={self.ratio}, " + \
            f"probability={self.probability}, random_attempts={self.random_attempts}, " + \
            f"min_roi_iou={self.min_roi_iou}, flags={self.flags})"

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        np.random.seed(seed)


class Resize:
    """Resizes a given image using OpenCV and numpy.

    This operation is deterministic. The code relies on OpenCV, meaning the interpolation arguments
    must be compatible with ``cv2.resize``.

    Attributes:
        interp: interpolation type to use (forwarded to ``cv2.resize``)
        buffer: specifies whether a destination buffer should be used to avoid allocations
        dsize: target image size (tuple of width, height).
        dst: buffer used to avoid reallocations if ``self.buffer == True``
        fx: argument forwarded to ``cv2.resize``.
        fy: argument forwarded to ``cv2.resize``.
    """

    def __init__(self, dsize, fx=0, fy=0, interp=cv.INTER_LINEAR, buffer=False):
        """Validates and initializes resize parameters.

        Args:
            dsize: size of the target image, forwarded to ``cv2.resize``.
            fx: x-scaling factor, forwarded to ``cv2.resize``.
            fy: y-scaling factor, forwarded to ``cv2.resize``.
            interp: resize interpolation type, forwarded to ``cv2.resize``.
            buffer: specifies whether a destination buffer should be used to avoid allocations
        """
        self.interp = thelper.utils.import_class(interp) if isinstance(interp, str) else interp
        self.buffer = buffer
        if isinstance(dsize, tuple):
            self.dsize = dsize
        else:
            self.dsize = tuple(dsize)
        self.dst = None
        self.fx = fx
        self.fy = fy
        assert 0 <= fx and 0 <= fy, "scale factors should be null (ignored) or positive"
        assert 0 <= dsize[0] and 0 <= dsize[1], "destination image size should be null (ignored) or positive"
        assert fx != 0 or fy != 0 or dsize[0] != 0 or dsize[1] != 0, \
            "need to specify either destination size or scale factor(s)"

    def __call__(self, sample):
        """Returns a resized copy of the provided image.

        Args:
            sample: the image to resize; should be a 2d or 3d numpy array.

        Returns:
            The resized image. May be allocated on the spot, or be a pointer to a local buffer.
        """
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        assert 2 <= sample.ndim <= 3, "bad input dimensions; must be 2-d, or 3-d (with channels)"
        if sample.ndim < 3 or sample.shape[2] <= 4:
            if self.buffer:
                cv.resize(sample, self.dsize, dst=self.dst, fx=self.fx, fy=self.fy, interpolation=self.interp)
                if self.dst.ndim == 2:
                    return np.expand_dims(self.dst, 2)
                return self.dst
            else:
                dst = cv.resize(sample, self.dsize, fx=self.fx, fy=self.fy, interpolation=self.interp)
                if dst.ndim == 2:
                    dst = np.expand_dims(dst, 2)
                return dst
        else:  # too many channels, need to split-resize
            slices_dst = self.dst if self.buffer else None
            slices = np.split(sample, sample.shape[2], 2)
            if slices_dst is None or not isinstance(slices_dst, list) or len(slices_dst) != len(slices):
                slices_dst = [None] * len(slices)
            for idx in range(len(slices)):
                slices_dst[idx] = cv.resize(slices[idx], self.dsize, dst=slices_dst[idx],
                                            fx=self.fx, fy=self.fy, interpolation=self.interp)
            if self.buffer:
                self.dst = slices_dst
            return np.stack(slices_dst, 2)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as data loss is incurred during image transformation."""
        # todo, could implement if original size is fixed & known
        raise NotImplementedError

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(dsize={self.dsize}, fx={self.fx}, fy={self.fy}, interp={self.interp}, buffer={self.buffer})"


class Affine:
    """Warps a given image using an affine matrix via OpenCV and numpy.

    This operation is deterministic. The code relies on OpenCV, meaning the border arguments
    must be compatible with ``cv2.warpAffine``.

    Attributes:
        transf: the 2x3 transformation matrix passed to ``cv2.warpAffine``.
        out_size: target image size (tuple of width, height). If None, same as original.
        flags: extra warp flags forwarded to ``cv2.warpAffine``.
        border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
        border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
    """

    def __init__(self, transf, out_size=None, flags=None, border_mode=None, border_val=None):
        """Validates and initializes affine warp parameters.

        Args:
            transf: the 2x3 transformation matrix passed to ``cv2.warpAffine``.
            out_size: target image size (tuple of width, height). If None, same as original.
            flags: extra warp flags forwarded to ``cv2.warpAffine``.
            border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
            border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
        """
        if isinstance(transf, np.ndarray):
            assert transf.size == 6, "transformation matrix must be 2x3"
            self.transf = transf.reshape((2, 3)).astype(np.float32)
        elif isinstance(transf, list):
            assert len(transf) == 6, "transformation matrix must be 6 elements (2x3)"
            self.transf = np.asarray(transf).reshape((2, 3)).astype(np.float32)
        else:
            raise TypeError("unexpected transformation matrix type")
        self.out_size = None
        if out_size is not None:
            if isinstance(out_size, list):
                assert len(out_size) == 2, "output image size should be 2-elem list or tuple"
                self.out_size = tuple(out_size)
            elif isinstance(out_size, tuple):
                assert len(out_size) == 2, "output image size should be 2-elem list or tuple"
                self.out_size = out_size
            else:
                raise TypeError("unexpected output size type")
        self.flags = flags
        if self.flags is None:
            self.flags = cv.INTER_LINEAR
        self.border_mode = border_mode
        if self.border_mode is None:
            self.border_mode = cv.BORDER_CONSTANT
        self.border_val = border_val
        if self.border_val is None:
            self.border_val = 0

    def __call__(self, sample):
        """Warps a given image using an affine matrix.

        Args:
            sample: the image to warp; should be a 2d or 3d numpy array.

        Returns:
            The warped image.
        """
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        out_size = self.out_size
        if out_size is None:
            out_size = (sample.shape[1], sample.shape[0])
        return cv.warpAffine(sample, self.transf, dsize=out_size, flags=self.flags,
                             borderMode=self.border_mode, borderValue=self.border_val)

    def invert(self, sample):
        """Inverts the warp transformation, but only is the output image has not been cropped before."""
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        out_size = self.out_size
        if out_size is None:
            out_size = (sample.shape[1], sample.shape[0])
        else:
            raise RuntimeError("unknown original image size, cannot invert affine transform")
        return cv.warpAffine(sample, self.transf, dsize=out_size, flags=self.flags ^ cv.WARP_INVERSE_MAP,
                             borderMode=self.border_mode, borderValue=self.border_val)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(transf={self.transf}, out_size={self.out_size}, flags={self.flags}, " + \
            f"border_mode={self.border_mode}, border_val={self.border_val})"


class RandomShift:
    """Randomly translates an image in a provided range via OpenCV and numpy.

    This operation is stochastic, and thus cannot be inverted. Each time the operation is called,
    a random check will determine whether a transformation is applied or not. The code relies on
    OpenCV, meaning the border arguments must be compatible with ``cv2.warpAffine``.

    Attributes:
        min: the minimum pixel shift that can be applied stochastically.
        max: the maximum pixel shift that can be applied stochastically.
        probability: the probability that the transformation will be applied when called.
        flags: extra warp flags forwarded to ``cv2.warpAffine``.
        border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
        border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
    """

    def __init__(self, min, max, probability=1.0, flags=None, border_mode=None, border_val=None):
        """Validates and initializes shift parameters.

        Args:
            min: the minimum pixel shift that can be applied stochastically.
            max: the maximum pixel shift that can be applied stochastically.
            probability: the probability that the transformation will be applied when called.
            flags: extra warp flags forwarded to ``cv2.warpAffine``.
            border_mode: border extrapolation mode forwarded to ``cv2.warpAffine``.
            border_val: border constant extrapolation value forwarded to ``cv2.warpAffine``.
        """
        if isinstance(min, tuple) and isinstance(max, tuple):
            assert len(min) == len(max) and len(min) == 2, "min/max shift tuple must be 2-elem"
            self.min = min
            self.max = max
        elif isinstance(min, list) and isinstance(max, list):
            assert len(min) == len(max) and len(min) == 2, "min/max shift list must be 2-elem"
            self.min = tuple(min)
            self.max = tuple(max)
        elif isinstance(min, (int, float)) and isinstance(max, (int, float)):
            self.min = (min, min)
            self.max = (max, max)
        else:
            raise TypeError("unexpected min/max combo types")
        assert self.max[0] >= self.min[0] or self.max[1] >= self.min[1], "bad min/max values"
        assert 0 <= probability <= 1, "bad probability range"
        self.probability = probability
        self.flags = flags
        if self.flags is None:
            self.flags = cv.INTER_LINEAR
        self.border_mode = border_mode
        if self.border_mode is None:
            self.border_mode = cv.BORDER_CONSTANT
        self.border_val = border_val
        if self.border_val is None:
            self.border_val = 0

    def __call__(self, sample):
        """Translates a given image using a predetermined min/max range.

        Args:
            sample: the image to translate; should be a 2d or 3d numpy array.

        Returns:
            The translated image.
        """
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        if self.probability < 1 and np.random.uniform(0, 1) > self.probability:
            return sample
        out_size = (sample.shape[1], sample.shape[0])
        x_shift = np.random.uniform(self.min[0], self.max[0])
        y_shift = np.random.uniform(self.min[1], self.max[1])
        transf = np.float32([[1, 0, x_shift], [0, 1, y_shift]])
        return cv.warpAffine(sample, transf, dsize=out_size, flags=self.flags, borderMode=self.border_mode, borderValue=self.border_val)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as it is stochastic, and data loss occurs during transformation."""
        raise RuntimeError("operation cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(min={self.min}, max={self.max}, probability={self.probability}, flags={self.flags}, " + \
            f"border_mode={self.border_mode}, border_val={self.border_val})"

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        np.random.seed(seed)


class ToGray:
    """Converts a multi-channel image to grayscale.

    This operation is deterministic, but not reversible. It can be applied to images with
    more than three channels (RGB) --- in that case, it will compute their per-pixel mean
    value. Note that in any case, the last dimension (that corresponds to the channels) will
    remain and be of size 1.
    """

    def __init__(self):
        """Does nothing, there's no attribute to store for this operation."""
        pass

    def __call__(self, sample):
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        assert 2 <= sample.ndim <= 3, "array should have at least two dimensions + channels"
        if sample.ndim == 2:
            return np.expand_dims(sample, 2)
        elif sample.shape[2] == 1:
            return sample  # already grayscale, return immediately
        elif sample.shape[2] == 3:
            return np.expand_dims(cv.cvtColor(sample, cv.COLOR_BGR2GRAY), 2)
        return np.mean(sample, axis=2, keepdims=True)

    def invert(self, sample):
        """Specifies that this operation cannot be inverted, as data loss occurs during transformation."""
        raise RuntimeError("operation cannot be inverted")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + "()"


class ToColor:
    """Converts a single-channel image to color (RGB).

    This operation is deterministic and reversible. It CANNOT be applied to images with
    more than one channel (HxWx1). The byte ordering (BGR or RGB) does not matter.
    """

    def __init__(self):
        """Does nothing, there's no attribute to store for this operation."""
        pass

    def __call__(self, sample):
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        assert 2 <= sample.ndim <= 3, "array should have at least two dimensions + channels"
        if sample.ndim == 2:
            return cv.cvtColor(sample, cv.COLOR_GRAY2BGR)
        elif sample.shape[2] == 1:
            return cv.cvtColor(sample[..., 0], cv.COLOR_GRAY2BGR)
        else:
            raise AssertionError("unexpected channel count in input sample")

    def invert(self, sample):
        """Inverts the operation by calling the 'ToGray' operation.

        Note that this operation is probably lossy due to OpenCV's grayscale conversion code
        which uses "0.21 R + 0.72 G + 0.07 B" to compute the luminosity of a pixel.
        """
        return ToGray()(sample)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + "()"


class Transpose:
    """Transposes an image via numpy.

    This operation is deterministic.

    Attributes:
        axes: the axes on which to apply the transpose; forwarded to ``numpy.transpose``.
        axes_inv: used to invert the tranpose; also forwarded to ``numpy.transpose``.
    """

    def __init__(self, axes):
        """Validates and initializes tranpose parameters.

        Args:
            axes: the axes on which to apply the transpose; forwarded to ``numpy.transpose``.
        """
        axes = np.asarray(axes)
        assert axes.ndim == 1, "tranpose param should be 1-d"
        assert not np.any(axes >= axes.size), "oob dim in axes"
        self.axes = axes
        self.axes_inv = np.asarray([None] * axes.size)
        for i in list(range(len(axes))):
            self.axes_inv[self.axes[i]] = i

    def __call__(self, sample):
        """Transposes a given image.

        Args:
            sample: the image to transpose; should be a numpy array.

        Returns:
            The transposed image.
        """
        assert isinstance(sample, np.ndarray), f"sample type should be np.ndarray (got {type(sample)})"
        return np.transpose(sample, self.axes)

    def invert(self, sample):
        """Invert-transposes a given image.

        Args:
            sample: the image to invert-transpose; should be a numpy array.

        Returns:
            The invert-transposed image.
        """
        return np.transpose(sample, self.axes_inv)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + f"(axes={self.axes})"


class Unsqueeze:
    """Expands a dimension in the input array via numpy/PyTorch.

    This operation is deterministic.

    Attributes:
        axis: the axis on which to apply the expansion.
    """

    def __init__(self, axis):
        """Validates and initializes tranpose parameters.

        Args:
            axis: the axis on which to apply the expansion.
        """
        self.axis = axis

    def __call__(self, sample):
        """Expands a dimension in the input array via numpy/PyTorch.

        Args:
            sample: the array to expand.

        Returns:
            The array with an extra dimension.
        """
        if isinstance(sample, np.ndarray):
            return np.expand_dims(sample, self.axis)
        elif isinstance(sample, torch.Tensor):
            return torch.unsqueeze(sample, self.axis)
        else:
            raise TypeError(f"unexpected input type ('{type(sample)}')")

    def invert(self, sample):
        """Squeezes a dimension in the input array via numpy/PyTorch.

        Args:
            sample: the array to squeeze.

        Returns:
            The array with one less dimension.
        """
        if isinstance(sample, np.ndarray):
            return np.squeeze(sample, self.axis)
        elif isinstance(sample, torch.Tensor):
            return torch.squeeze(sample, self.axis)
        else:
            raise TypeError(f"unexpected input type ('{type(sample)}')")

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + f"(axis={self.axis})"


class Duplicator(NoTransform):
    """Duplicates and returns a list of copies of the input sample.

    This operation is used in data augmentation pipelines that rely on probabilistic or preset transformations.
    It can produce a fixed number of simple copies or deep copies of the input samples as required.

    .. warning::
        Since the duplicates will be given directly to the data loader as part of the same minibatch, using too
        many copies can adversely affect gradient descent for that minibatch. To simply increase the total size
        of the training set while still allowing a proper shuffling of samples and/or to keep the minibatch size
        intact, we instead recommend setting the ``train_scale`` configuration value in the data loader. See
        :func:`thelper.data.utils.create_loaders` for more information.

    Attributes:
        count: number of copies to generate.
        deepcopy: specifies whether to deep-copy samples or not.
    """

    def __init__(self, count, deepcopy=False):
        """Validates and initializes duplication parameters.

        Args:
            count: number of copies to generate.
            deepcopy: specifies whether to deep-copy samples or not.
        """
        assert count > 0, "invalid copy count"
        self.count = count
        self.deepcopy = deepcopy

    def __call__(self, sample):
        """Generates and returns duplicates of the sample/object.

        If a dictionary is provided, its values will be expanded into lists that
        will contain all duplicates. Otherwise, the duplicates will be returned
        directly as a list.

        Args:
            sample: the sample/object to duplicate.

        Returns:
            A list of duplicated samples, or a dictionary of duplicate lists.
        """
        copyfct = copy.deepcopy if self.deepcopy else copy.copy
        if isinstance(sample, dict):
            return {k: [copyfct(v) for _ in range(self.count)] for k, v in sample.items()}
        else:
            return [copyfct(sample) for _ in range(self.count)]

    def invert(self, sample):
        """Returns the first instance of the list of duplicates."""
        if isinstance(sample, dict):
            assert all([isinstance(v, list) for v in sample.values()]), "invalid sample type (should be list)"
            assert not any([len(v) != self.count for v in sample.values()]), "invalid sample list length"
            return {k: v[0] for k, v in sample.items()}
        else:
            assert isinstance(sample, list), "invalid sample type (should be list)"
            assert len(sample) == self.count, "invalid sample list length"
            return sample[0]

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(count={self.count}, deepcopy={self.deepcopy})"


class Tile:
    """Returns a list of tiles cut out from a given image.

    This operation can perform tiling given an optional mask with a target intersection over union (IoU) score,
    and with an optional overlap between tiles. The tiling is deterministic and can thus be inverted, but only
    if a mask is not used, as some image regions may be lost otherwise.

    If a mask is used, the first tile position is tested exhaustively by iterating over all input coordinates
    starting from the top-left corner of the image. Otherwise, the first tile position is set as (0,0). Then,
    all other tiles are found by offsetting frm these coordinates, and testing for IoU with the mask (if needed).

    Attributes:
        tile_size: size of the output tiles, provided as a single element (``edge_size``) or as a
            two-element tuple or list (``[width, height]``). If integer values are used, the size is
            assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
            size is assumed to be relative to the original image size, and will be determined at
            execution time for each image.
        tile_overlap: overlap allowed between two neighboring tiles; should be a ratio in [0,1].
        min_mask_iou: minimum mask intersection over union (IoU) required for accepting a tile (in [0,1]).
        offset_overlap: specifies whether the overlap tiling should be offset outside the image or not.
        bordertype: border copy type to use when the image is too small for the required crop size.
            See ``cv2.copyMakeBorder`` for more information.
        borderval: border value to use when the image is too small for the required crop size. See
            ``cv2.copyMakeBorder`` for more information.
    """

    def __init__(self, tile_size, tile_overlap=0.0, min_mask_iou=1.0, offset_overlap=False, bordertype=cv.BORDER_CONSTANT, borderval=0):
        """Validates and initializes tiling parameters.

        Args:
            tile_size: size of the output tiles, provided as a single element (``edge_size``) or as a
                two-element tuple or list (``[width, height]``). If integer values are used, the size is
                assumed to be absolute. If floating point values are used (i.e. in [0,1]), the output
                size is assumed to be relative to the original image size, and will be determined at
                execution time for each image.
            tile_overlap: overlap ratio between two consecutive (neighboring) tiles; should be in [0,1].
            min_mask_iou: minimum mask intersection over union (IoU) required for producing a tile.
            offset_overlap: specifies whether the overlap tiling should be offset outside the image or not.
            bordertype: border copy type to use when the image is too small for the required crop size.
                See ``cv2.copyMakeBorder`` for more information.
            borderval: border value to use when the image is too small for the required crop size. See
                ``cv2.copyMakeBorder`` for more information.
        """
        if isinstance(tile_size, (tuple, list)):
            assert len(tile_size) == 2, "expected tile size to be two-element list or tuple, or single scalar"
            assert all([isinstance(s, int) for s in tile_size]) or all([isinstance(s, float) for s in tile_size]), \
                "expected tile size pair elements to be the same type (int or float)"
            self.tile_size = tile_size
        elif isinstance(tile_size, (int, float)):
            self.tile_size = (tile_size, tile_size)
        else:
            raise TypeError("unexpected tile size type (need tuple/list/int/float)")
        assert isinstance(tile_overlap, float) and 0 <= tile_overlap < 1, \
            "invalid tile overlap (should be float in [0,1[)"
        self.tile_overlap = tile_overlap
        assert isinstance(min_mask_iou, float) and 0 <= min_mask_iou <= 1, \
            "invalid minimum mask IoU score (should be float in [0,1])"
        self.min_mask_iou = min_mask_iou
        self.offset_overlap = offset_overlap
        self.bordertype = thelper.utils.import_class(bordertype) if isinstance(bordertype, str) else bordertype
        self.borderval = borderval

    def __call__(self, image, mask=None):
        """Extracts and returns a list of tiles cut out from the given image.

        Args:
            image: the image to cut into tiles. If given as a 2-element list, it is assumed to contain both
                the image and the mask (passed through a composer).
            mask: the mask to check tile intersections with (may be ``None``).

        Returns:
            A list of tiles (numpy-compatible images).
        """
        if isinstance(image, list) and len(image) == 2:
            assert mask is None, "mask provided twice"
            # we assume that the mask was given as the 2nd element of the list
            image, mask = image[0], image[1]
        tile_rects, tile_images = self._get_tile_rects(image, mask), []
        for rect in tile_rects:
            tile_images.append(thelper.draw.safe_crop(image, (rect[0], rect[1]),
                                                      (rect[0] + rect[2], rect[1] + rect[3]),
                                                      self.bordertype, self.borderval))
        return tile_images

    def count_tiles(self, image, mask=None):
        """Returns the number of tiles that would be cut out from the given image.

        Args:
            image: the image to cut into tiles. If given as a 2-element list, it is assumed to contain both
                the image and the mask (passed through a composer).
            mask: the mask to check tile intersections with (may be ``None``).

        Returns:
            The number of tiles that would be cut with :func:`thelper.transforms.operations.Tile.__call__`.
        """
        if isinstance(image, list) and len(image) == 2:
            assert mask is None, "mask provided twice"
            # we assume that the mask was given as the 2nd element of the list
            image, mask = image[0], image[1]
        return len(self._get_tile_rects(image, mask))

    def _get_tile_rects(self, image, mask=None):
        assert isinstance(image, (PIL.Image.Image, np.ndarray)), \
            "image type should be np.ndarray or PIL image"
        if isinstance(image, PIL.Image.Image):
            image = np.asarray(image)
        if mask is not None:
            assert isinstance(mask, (PIL.Image.Image, np.ndarray)), \
                "mask type should be np.ndarray or PIL image"
            if isinstance(mask, PIL.Image.Image):
                mask = np.asarray(mask)
        tile_rects = []
        height, width = image.shape[0], image.shape[1]
        if isinstance(self.tile_size[0], float):
            tile_size = (int(round(self.tile_size[0] * width)), int(round(self.tile_size[1] * height)))
        else:
            tile_size = self.tile_size
        overlap = (int(round(tile_size[0] * self.tile_overlap)), int(round(tile_size[1] * self.tile_overlap)))
        overlap_offset = (-overlap[0] // 2, -overlap[1] // 2) if self.offset_overlap else (0, 0)
        step_size = (max(tile_size[0] - (overlap[0] // 2) * 2, 1), max(tile_size[1] - (overlap[1] // 2) * 2, 1))
        req_mask_area = tile_size[0] * tile_size[1] * self.min_mask_iou
        if mask is not None:
            assert height == mask.shape[0] and width == mask.shape[1], "image and mask dimensions mismatch"
            assert mask.ndim == 2, "mask should be 2d binary (uchar) array"
            offset_coord = None
            row_range = range(overlap_offset[1], height - overlap_offset[1] - tile_size[1] + 1)
            col_range = range(overlap_offset[0], width - overlap_offset[0] - tile_size[0] + 1)
            for row, col in itertools.product(row_range, col_range):
                crop = thelper.draw.safe_crop(mask, (col, row), (col + tile_size[0], row + tile_size[1]))
                if np.count_nonzero(crop) >= req_mask_area:
                    offset_coord = (overlap_offset[0] + ((col - overlap_offset[0]) % step_size[0]),
                                    overlap_offset[1] + ((row - overlap_offset[1]) % step_size[1]))
                    break
            if offset_coord is None:
                return tile_rects
        else:
            offset_coord = overlap_offset
        row = offset_coord[1]
        while row + tile_size[1] <= height - overlap_offset[1]:
            col = offset_coord[0]
            while col + tile_size[0] <= width - overlap_offset[0]:
                if mask is not None:
                    crop = thelper.draw.safe_crop(mask, (col, row), (col + tile_size[0], row + tile_size[1]))
                    if np.count_nonzero(crop) >= req_mask_area:
                        tile_rects.append((col, row, tile_size[0], tile_size[1]))  # rect = (x, y, w, h)
                else:
                    tile_rects.append((col, row, tile_size[0], tile_size[1]))  # rect = (x, y, w, h)
                col += step_size[0]
            row += step_size[1]
        return tile_rects

    def invert(self, image, mask=None):
        """Returns the reconstituted image from a list of tiles, or throws if a mask was used."""
        assert mask is None, "cannot invert operation, mask might have forced the loss of image content"
        raise NotImplementedError  # TODO

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(tile_size={self.tile_size}, tile_overlap={self.tile_overlap}, min_mask_iou={self.min_mask_iou}, " + \
            f"offset_overlap={self.offset_overlap}, bordertype={self.bordertype}, borderval={self.borderval})"


class NormalizeZeroMeanUnitVar:
    """Normalizes a given image using a set of mean and standard deviation parameters.

    The samples will be transformed such that ``s = (s - mean) / std``.

    This can be used for whitening; see https://en.wikipedia.org/wiki/Whitening_transformation
    for more information. Note that this operation is also not restricted to images.

    Attributes:
        mean: an array of mean values to subtract from data samples.
        std: an array of standard deviation values to divide with.
        out_type: the output data type to cast the normalization result to.
    """

    def __init__(self, mean, std, out_type=np.float32):
        """Validates and initializes normalization parameters.

        Args:
            mean: an array of mean values to subtract from data samples.
            std: an array of standard deviation values to divide with.
            out_type: the output data type to cast the normalization result to.
        """
        self.out_type = out_type
        self.mean = np.asarray(mean).astype(out_type)
        self.std = np.asarray(std).astype(out_type)
        assert self.mean.ndim == 1 and self.std.ndim == 1, "normalization params should be 1-d"
        assert self.mean.size == self.std.size, "normalization params size mismatch"
        assert not any([d == 0 for d in self.std]), "normalization std must be non-null"

    def __call__(self, sample):
        """Normalizes a given sample.

        Args:
            sample: the sample to normalize. If given as a PIL image, it
            will be converted to a numpy array first.

        Returns:
            The warped sample, in a numpy array of type ``self.out_type``.
        """
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        if isinstance(sample, np.ndarray):
            return ((sample - self.mean) / self.std).astype(self.out_type)
        elif isinstance(sample, torch.Tensor):
            out = torchvision.transforms.functional.normalize(sample,
                                                              torch.from_numpy(self.mean),
                                                              torch.from_numpy(self.std))
            assert self.out_type != np.float32, "missing impl for non-float torch normalize output"
            return out.float()
        else:
            raise TypeError(f"sample type should be np.ndarray or torch.Tensor (got {type(sample)})")

    def invert(self, sample):
        """Inverts the normalization."""
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        return (sample * self.std + self.mean).astype(self.out_type)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(mean={self.mean}, std={self.std}, out_type={self.out_type})"


class NormalizeMinMax:
    """Normalizes a given image using a set of minimum and maximum values.

    The samples will be transformed such that ``s = (s - min) / (max - min)``.

    Note that this operation is also not restricted to images.

    Attributes:
        min: an array of minimum values to subtract with.
        max: an array of maximum values to divide with.
        out_type: the output data type to cast the normalization result to.
    """

    def __init__(self, min, max, out_type=np.float32):
        """Validates and initializes normalization parameters.

        Args:
            min: an array of minimum values to subtract with.
            max: an array of maximum values to divide with.
            out_type: the output data type to cast the normalization result to.
        """
        self.out_type = out_type
        self.min = np.asarray(min).astype(out_type)
        if self.min.ndim == 0:
            self.min = np.expand_dims(self.min, 0)
        self.max = np.asarray(max).astype(out_type)
        if self.max.ndim == 0:
            self.max = np.expand_dims(self.max, 0)
        assert self.min.ndim == 1 and self.max.ndim == 1, \
            "normalization params should be a 1-d array (one value per channel)"
        assert self.min.size == self.max.size, "normalization params size mismatch"
        self.diff = self.max - self.min
        assert not any([d == 0 for d in self.diff]), "normalization diff must be non-null"

    def __call__(self, sample):
        """Normalizes a given sample.

        Args:
            sample: the sample to normalize. If given as a PIL image, it
            will be converted to a numpy array first.

        Returns:
            The warped sample, in a numpy array of type ``self.out_type``.
        """
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        return ((sample - self.min) / self.diff).astype(self.out_type)

    def invert(self, sample):
        """Inverts the normalization."""
        assert isinstance(sample, (PIL.Image.Image, np.ndarray)), \
            f"sample type should be np.ndarray or PIL image (got {type(sample)})"
        if isinstance(sample, PIL.Image.Image):
            sample = np.asarray(sample)
        return (sample * self.diff + self.min).astype(self.out_type)

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(min={self.min}, max={self.max}, out_type={self.out_type})"
